/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.backup.plugins;

import de.bsvrz.ars.ars.backup.BackupException;
import de.bsvrz.ars.ars.backup.BackupModul;
import de.bsvrz.ars.ars.backup.MediumNotAccessibleException;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.SuspendableTask;
import de.bsvrz.ars.ars.persistence.ContainerFile;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.*;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;

/**
 * Diese Klasse ist fr die Sicherung von Containern auf dem Filesystem verantwortlich.
 *
 * @author beck et al. projects GmbH
 * @author Christian Wied
 * @version $Revision: 8347 $ / $Date: 2010-11-25 11:06:58 +0100 (Thu, 25 Nov 2010) $ / ($Author: jh $)
 */
public class FileSystemBackup implements BackupModul {

	public static final String PROP_BACKUP_DIR = "sicherungsVerzeichnis";

	protected static int MAX_CONTAINERS_PER_DIR = 1000;

	protected static long MAX_MEDIUM_SIZE_KB = 4300L * 1024L;

	protected Debug logger = Debug.getLogger();

	private int currentContIndex = 0;

	private List<File> currentContainersToBeSaved = new ArrayList<File>();

	protected long currentSpaceOccupiedByZipfiles = 0;

	protected long currentSpaceOccupiedByContainerFiles = 0;

	protected String currentBackupPath;

	protected String backupBasePath = null;

	public FileSystemBackup() {
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#initialize(java.util.Properties) */
	public void initialize(Properties props) throws Exception {

		String zipSize = props.getProperty(PROP_BACKUP_CONTAINER_PER_ZIP);
		try {
			if(zipSize != null) {
				MAX_CONTAINERS_PER_DIR = Integer.parseInt(zipSize);
			}
		}
		catch(NumberFormatException e) {
		} // do nothing, default 100 is ok.

		String mediumSize = props.getProperty(PROP_BACKUP_MEDIUMSIZE);
		try {
			if(mediumSize != null) {
				MAX_MEDIUM_SIZE_KB = Long.parseLong(mediumSize);
			}
		}
		catch(NumberFormatException e) {
		} // do nothing, take default.

		String backupDirName = props.getProperty(PROP_BACKUP_DIR);

		if(backupDirName == null) {
			throw new Exception("Bentigter Parameter '" + PROP_BACKUP_DIR + "' nicht angegeben.");
		}
		File backupDir = new File(backupDirName);
		if(!backupDir.exists()) {
			backupDir.mkdir();
		}
		backupBasePath = backupDir.getAbsolutePath();
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#hasMediumCapacity(int,long,int) */
	public boolean hasMediumCapacity(int mediumID, long containerSize, int mediumIndexSize) throws BackupException {
		return MAX_MEDIUM_SIZE_KB * 1024L > currentSpaceOccupiedByZipfiles + currentSpaceOccupiedByContainerFiles + containerSize + mediumIndexSize;
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#openMediumBeforeBackup(int,java.lang.String) */
	public void openMediumBeforeBackup(int mediumID, String backupRunID) throws BackupException {
		String tmpMediumPath = backupBasePath + File.separator + "Medium_" + mediumID;
		if(!new File(tmpMediumPath).exists()) {
			new File(tmpMediumPath).mkdir();
		}
		currentBackupPath = tmpMediumPath + File.separator + backupRunID;
		if(!new File(currentBackupPath).exists()) {
			new File(currentBackupPath).mkdir();
		}
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#backupContainer(int,long,java.io.File) */
	public void backupContainer(int mediumID, long contID, File contFile) throws BackupException {

		try {
			// bis max-Grenze erreicht ist: File merken
			if(++currentContIndex % MAX_CONTAINERS_PER_DIR != 0) {
				currentContainersToBeSaved.add(contFile);
				currentSpaceOccupiedByContainerFiles += contFile.length();
			} // wenn max-Grenze erreicht ist: Zippen, Speichern und neues Verzeichnis bestimmen
			else {
				currentContainersToBeSaved.add(contFile);
				currentSpaceOccupiedByContainerFiles += contFile.length();
				backupContainerList();
			}
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#closeMediumAfterBackup(int,java.io.File) */
	@SuppressWarnings("unchecked")
	public void closeMediumAfterBackup(int mediumID, File indexFile) throws BackupException {
		try {
			if(currentContainersToBeSaved.size() != 0) {
				// die letzten Container noch zippen und schreiben:
				backupContainerList();
			}

			// indexFile abspeichern
			File newIndexFile = new File(currentBackupPath, indexFile.getName());
			FileInputStream is = new FileInputStream(indexFile);
			FileOutputStream os = new FileOutputStream(newIndexFile);
			copyStream(is, os);
			os.close();
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
		finally {
			currentContainersToBeSaved.clear();
			currentSpaceOccupiedByContainerFiles = 0;
			currentSpaceOccupiedByZipfiles = 0;
		}
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#restoreContainer(int,java.lang.String) */
	public InputStream restoreContainer(int mediumID, String contFileName) throws BackupException {

		String backupPath = backupBasePath;
		if(!backupBasePath.startsWith(DVDBackup.MEDIAPATH_PREFIX)) {
			// Bei DVD gibt's das Medium-Verzeichnis nicht - daher bei filesystem anhaengen: 
			backupPath += File.separator + "Medium_" + mediumID;
		}
		logger.fine("Suche nach Container " + contFileName + " im Verzeichnis: " + backupPath);

		long contID = ContainerFile.getContID(contFileName);
		File zipFile = null;
		try {
			List<File> backupRuns = Arrays.asList(Util.listDirectories(backupPath));
			Collections.sort(backupRuns);

			// indexFile auspacken und nachschauen, ob Container enthalten ist
			boolean containerFound = false;
			int i = backupRuns.size() - 1;
			String backupRun = "";
			while(!containerFound && i >= 0) {
				backupRun = backupRuns.get(i--).getName();

				File indexFile = new File(backupPath + File.separator + backupRun, BackupTask.INDEXFILE_NAME);
				if(!indexFile.exists()) {
					logger.info("kein Index-File gefunden: " + indexFile.getAbsolutePath());
					continue;
				}
				
				RandomAccessFile raf = new RandomAccessFile(indexFile, "r");
				for(String line; (line = raf.readLine()) != null;) {
					if(line.indexOf(contFileName) != -1) {
						containerFound = true;
						break;
					}
				}
				raf.close();
			}
			if(!containerFound) {
				logger.error("Container " + contFileName + " konnte nicht auf MediumId=" + mediumID + " gefunden werden.");
				throw new BackupException("Container " + contFileName + " konnte nicht auf MediumId=" + mediumID + " gefunden werden.");
			}

			// nach dem richtigen Zip-File suchen:
			File[] zipfiles = new File(backupPath, backupRun).listFiles(
					new FileFilter() {

						public boolean accept(File pathname) {
							return pathname.getName().endsWith(".zip");
						}
					}
			);

			for(i = 0; i < zipfiles.length; i++) {
				String name = zipfiles[i].getName();
				String firstContainerNameAbb = name.substring(0, name.indexOf("-"));
				String lastContainerNameAbb = name.substring(name.indexOf("-") + 1, name.length() - 4);
				long firstContID = Long.parseLong(firstContainerNameAbb.substring(2, firstContainerNameAbb.length()));
				long lastContID = Long.parseLong(lastContainerNameAbb.substring(2, lastContainerNameAbb.length()));

				if(contID >= firstContID && contID <= lastContID) {
					zipFile = new File(backupPath + File.separator + backupRun, name);
					break;
				}
			}
			if(zipFile == null) {
				logger.warning("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.");
				throw new BackupException("Zip-Datei mit ContainerId=" + contID + " konnte nicht gefunden werden.");
			}
		}
		catch(Exception e) {
			logger.warning("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.", e.getMessage());
			throw new BackupException("Zip-Datei mit ContainerId=" + contID + " konnte nicht auf MediumID=" + mediumID + " gefunden werden.");
		}
		try {
			File file = detachFileFromZip(zipFile, contFileName);
			file.deleteOnExit();
			return new FileInputStream(file);
		}
		catch(Exception e) {
			throw new BackupException(e);
		}
	}

	/** @see de.bsvrz.ars.ars.backup.BackupModul#getAllContFileNames(SuspendableTask,int,int) */
	public List<String> getAllContFileNames(SuspendableTask task, int mediumID, int maxLimit) throws BackupException {

		String backupPath = backupBasePath + File.separator + "Medium_" + mediumID;

		File tmpBackupPathFile = new File(backupPath);
		if(!tmpBackupPathFile.exists()) {
			logger.error("BackupPath " + backupPath + " existiert nicht.");
			throw new MediumNotAccessibleException(mediumID);
		}

		List<String> result = new ArrayList<String>();
		List<File> backupRuns = Arrays.asList(Util.listDirectories(backupPath));
		Collections.sort(backupRuns);

		int counter = 1;
		String backupRun = "";
		for(int i = backupRuns.size() - 1; i >= 0; i--) {
			task.suspendTaskIfNecessary();	//Task anhalten, falls Archivsystem berlastet
			backupRun = backupRuns.get(i).getName();

			File indexFile = new File(backupPath + File.separator + backupRun, BackupTask.INDEXFILE_NAME);
			if(!indexFile.exists()) {
				continue;
			}
			try {
				RandomAccessFile raf = new RandomAccessFile(indexFile.getAbsolutePath(), "r");
				for(String line; (line = raf.readLine()) != null;) {
					result.add(line.substring(line.lastIndexOf(File.separator) + 1));
					if(maxLimit != -1 && counter++ >= maxLimit) {
						return result;
					}
				}
				raf.close();
			}
			catch(Exception e) {
				Debug.getLogger().error("FileSystemBackup: Indexdatei " + indexFile.getAbsolutePath() + " konnte nicht gelesen werden.", e.getMessage());
			}
		}
		return result;
	}

	/**
	 * Entpackt eine ContainerDatei aus einem Zip-File.
	 *
	 * @param zipFile
	 * @param contFileName
	 *
	 * @return File
	 *
	 * @throws Exception
	 */
	private File detachFileFromZip(File zipFile, String contFileName) throws Exception {

		File restoreFile = new File(contFileName);
		if(restoreFile.exists()) restoreFile.delete();
		restoreFile.createNewFile();
		// wir holen uns den entsprechenden Container aus dem Zip-File:
		try {
			ZipInputStream zin = new ZipInputStream(new BufferedInputStream(new FileInputStream(zipFile)));
			ZipEntry zen;
			while((zen = zin.getNextEntry()) != null) {
				if(zen.getName().equals(contFileName)) {
					OutputStream os = new BufferedOutputStream(new FileOutputStream(restoreFile));
					copyStream(zin, os);
					os.close();
					restoreFile.deleteOnExit();
					return restoreFile;
				}
			}
		}
		catch(IOException e) {
			throw new BackupException(e);
		}
		throw new BackupException(
				"Fehler beim Wiederherstellen: Datei '" + contFileName + "' konnte nicht aus Zip-Datei '" + zipFile.getAbsolutePath() + "' extrahiert werden."
		);
	}

	/**
	 * Zippt die in der Containerlist aufgelisteten Dateien und legt sie in den <code>currentBackupPath</code>.
	 *
	 * @throws Exception
	 */
	private void backupContainerList() throws Exception {

		File backupDir = new File(currentBackupPath);

		Collections.sort(currentContainersToBeSaved, new FileNameComparator());

		String nameFirst = currentContainersToBeSaved.get(0).getName();
		String nameLast = currentContainersToBeSaved.get(currentContainersToBeSaved.size() - 1).getName();
		String containerZipName = nameFirst.substring(0, nameFirst.indexOf("."));
		containerZipName += "-" + nameLast.substring(0, nameLast.indexOf("."));

		ZipOutputStream out;
		File zipFile = new File(backupDir, containerZipName + ".zip");
		zipFile.createNewFile();
		out = new ZipOutputStream(new FileOutputStream(zipFile));
		for(File file : currentContainersToBeSaved) {
			ZipEntry entry = new ZipEntry(file.getName());
			entry.setSize(file.length());
			entry.setTime(file.lastModified());
			out.putNextEntry(entry);
			copyStream(new FileInputStream(file), out);
			out.closeEntry();
		}
		out.finish();
		out.close();
		currentSpaceOccupiedByZipfiles += zipFile.length();
		currentSpaceOccupiedByContainerFiles = 0;
		currentContainersToBeSaved.clear();
	}

	private void copyStream(InputStream in, OutputStream out) throws IOException {
		byte[] buffer = new byte[1024];
		int len;
		while((len = in.read(buffer)) >= 0) out.write(buffer, 0, len);

		in.close();
		// out.close();
	}

	private class FileNameComparator implements Comparator<File> {

		public int compare(File f1, File f2) {
			return f1.getName().compareTo(f2.getName());
		}
	}

	public String toString() {
		return "FileSystem (" + backupBasePath + ")";
	}
}
