/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.ArchivConfig.SubscrMarker.SubscrInfo;
import de.bsvrz.ars.ars.mgmt.datatree.DIdNodeNotFoundException;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentNode;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTreeIterator;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationResultData;
import de.bsvrz.ars.ars.mgmt.tasks.ArchiveTask;
import de.bsvrz.ars.ars.mgmt.tasks.Task;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.dataIdentificationSettings.EndOfSettingsListener;
import de.bsvrz.sys.funclib.dataIdentificationSettings.SettingsManager;
import de.bsvrz.sys.funclib.dataIdentificationSettings.UpdateListener;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.ringbuffer.RingBuffer;
import de.bsvrz.sys.funclib.losb.util.K;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.*;

/**
 * Konfiguration des Archivsystems. Die Klasse ist als Thread implementiert, damit die Anmeldungen nicht im Thread des Datenverteilers (in den Callback-Methoden
 * des SettingsManager) laufen muessen. Stattdessen wird ein einfacher Ping-Pong-Mechnanismus implementiert, mit dem die Parameter abgearbeitet werden, sobald
 * der EndOfSettingsListener aufgerufen wurde. Das Ende eines Parametrierungszyklus kann durch {@link #createEndOfParamListener()} abgewartet werden.
 * <p/>
 * In einem Parametrierungszyklus werden alle An-/Abmeldungen von zu archivierenden Datenidentifikationen und Quittierungs-Datenidentifikationen vorgenommen.
 * <p/>
 * Diese Klasse ist als einzige nicht mit Hilfe von {@link Task} und {@link RingBuffer} implementiert, da sie einen von der Kernsoftware zur Verfuegung
 * gestellten Mechanismus zur Parameteruebernahme verwendet.
 *
 * @author beck et al. projects GmbH
 * @author Alexander Schmidt
 * @version $Revision: 6437 $ / $Date: 2009-03-26 16:36:14 +0100 (Thu, 26 Mar 2009) $ / ($Author: rs $)
 */
public class ArchivConfig extends Task {

	private static final String ARS_PARAM_ATG_PID = "atg.archiv";

	private static final String ARS_PARAM_ASP_PID = "asp.parameterSoll";

	private static final String DEFAULT_PARAMS_FILE = "ArchivsystemParameter.txt";

	/** Bei Ausgeben der Parametrierung wird diese Zeitspanne auf den Parameter-Datensatz gewartet */
	private static final long PARAM_TIMEOUT = 10000;

	private boolean shallTerminate = false, isTerminated = false;

	private Object terminationMutex = new Object();

	private boolean paramsAvailable = false;

	private InQueuesMgr inQueuesMgr = null;

	private int totalDIDs = 0;

	private SettingsManager settingsMgr = null;

	private EndOfSettingsListener eofSettingsLstnr;

	private UpdateListener updateSettingsLstnr;

	private SubscrMarker subscrMarker = null;

	private FileOutputStream paramDumpStream = null;

	private boolean printParams;

	private boolean noAnswer = false;

	private boolean gotFirstPrintParam = false;

	private RingBuffer<ResultData> simParamQueue;

	/** Listeners die benachrichtigt werden, wenn eine Parametrierung beendet ist. Diese Liste wird nach jedem Durchlauf geleert. */
	private List<EndOfParamListener> endOfParamListeners = new ArrayList<EndOfParamListener>();

	public ArchivConfig(ArchiveManager aMgr) throws ConfigurationException {
		super(aMgr, null);
		logger = Debug.getLogger();
		setName(getClass().getSimpleName());
	}

	public void setSimConfigQueue(RingBuffer<ResultData> queue) {
		simParamQueue = queue;
	}

	private SettingsManager createSettingsMgr() throws ConfigurationException {
		SettingsManager sMgr = new SettingsManager(
				archMgr.getDavCon(), new DataIdentification(
				archMgr.getConfigAuth(), archMgr.getDD(ARS_PARAM_ATG_PID, ARS_PARAM_ASP_PID)
		)
		);

		// Callback-Methode: Wird vom DAV aufgerufen, wenn eine an/abzumeldende DID identifiziert wurde
		updateSettingsLstnr = new UpdateListener() {
			public void update(DataIdentification did, Data oldD, Data newD) {
				try {
					// Falls das An/Abmelden noch nicht fertig ist, blockieren wir die update()-Methode ein wenig.
					synchronized(ArchivConfig.this) {
						while(paramsAvailable) {
							try {
								ArchivConfig.this.wait();
							}
							catch(InterruptedException e) {
								return;
							}
						}
					}
					totalDIDs++;
					if(printParams) {
						printParam(paramDumpStream, totalDIDs + "", did, newD);
					}
					else {
						// Erst Parametrierung bernehmen, dann an/abmelden. Dadurch kann man falls ntig
						// DIDs zu schwungweiser An-/Abmeldung zusammenfassen (wird derzeit nicht gemacht).

						if(oldD == null && newD != null)		// Fall 1: Neu parametriert
						{
							prepareNewParam(did, newD);
						}
						else if(oldD != null && newD == null)	// Fall 2: Nicht mehr parametriert
						{
							prepareUnparam(did);
						}
						else if(oldD != null && newD != null)	// Fall 3: Umparametriert
						{
							prepareReparam(did, newD);
						}
						else {
							logger.warning("Archivparametrierung: ungueltige Parameter (oldD==newD==null)");
						}
					}
				}
				catch(NullPointerException npe) {
					Debug.getLogger().warning("Fehler beim Auswerten der eingehenden Daten. Daten werden ignoriert! " + did + ", " + oldD + ", " + newD, npe);
				}
			}
		};
		sMgr.addUpdateListener(updateSettingsLstnr);

		// Callback-Methode: Wird vom DAV aufgerufen, alle An/Abzumeldungen beendet sind
		eofSettingsLstnr = new EndOfSettingsListener() {
			public void inform() {
				synchronized(ArchivConfig.this) {	// Vorsicht, anonymes Objekt!!
					paramsAvailable = true;		// der verarbeitende Thread soll loslaufen...
					ArchivConfig.this.notify();
				}
			}
		};
		sMgr.addEndOfSettingsListener(eofSettingsLstnr);
		return sMgr;
	}

	public void run() {
		while(!shallTerminate) {
			try {
				synchronized(this) {
					while(!paramsAvailable && (simParamQueue == null || simParamQueue.isEmpty()))	// wird im EndOfSettingsListener auf wahr gesetzt
					{
						wait();
					}
				}
				if(paramsAvailable) {
					processParams();
				}
				else {
					processSimParams(simParamQueue.pop());
				}

				synchronized(endOfParamListeners) {
					for(EndOfParamListener eopl : endOfParamListeners) {
						eopl.gotParams();
					}
					endOfParamListeners.clear();
				}
				paramsAvailable = false;

				synchronized(this) {
					notify();		// den ggf. wartenden UpdateListener weitermachen lassen
				}
			}
			catch(InterruptedException e) {
				break;
			}
		}
		synchronized(terminationMutex) {	// wartende Termination-Methode benachrichtigen
			isTerminated = true;
			terminationMutex.notify();
		}
		simParamQueue = null;
		endOfParamListeners = null;
		subscrMarker = null;
	}

	/**
	 * Meldet Daten einer Simulation zum Archivieren (und Quittieren) an. <code>rd</code> muss vom Typ {@link SimulationResultData} sein, sonst wird keine
	 * Anmeldung vorgenommen.
	 *
	 * @param rd Spezifikation der anzumeldenden Daten. Muss vom Typ {@link SimulationResultData} sein.
	 */
	private void processSimParams(ResultData rd) {
		if(rd instanceof SimulationResultData) {
			boolean successful = true;
			SimulationResultData ard = ((SimulationResultData)rd);
			Map<DataIdentification, Data> settingsMap = new HashMap<DataIdentification, Data>();
			SimVarParamExtracter.extractSettings(ard, archMgr.getDataModel(), settingsMap, ard.getSimVar());
			int i = 0;
			SubscrMarker sm = new SubscrMarker();	// zum Mitzaehlen der Quittierungen eigenen Marker anlegen
			for(DataIdentification did : settingsMap.keySet()) {
				i++;
				DataDescription dd = did.getDataDescription();
				DataIdentNode din;
				try {
					din = didTree.getDataIdentNode(did);
				}
				catch(DIdNodeNotFoundException e) {
					din = new DataIdentNode();
				}
				din.arsParamSetParams(settingsMap.get(did));				// Teil der aktuellen Parametrierung		
				didTree.putDataIdentNode(did, din);

				String logMsg = "";
				try {
					// Beim ersten DS nach Anmeldung darf der DIdx ausnahmsweise gleich dem letzten DIdx sein (wenn sich die DID nicht
					// geaendert hat). Um dies feststellen zu koennen, mussen folgende Flags gesetzt werden:
					din.setFirstDataAfterSubscription(ArchiveDataKind.ONLINE, true);
					din.setFirstDataAfterSubscription(ArchiveDataKind.ONLINE_DELAYED, true);

					// erst die Quittung, damit sofort quittiert werden kann
					if(din.arSParamIsQuittieren()) {
						final SystemObject so = did.getObject();
						final AttributeGroup atg = dd.getAttributeGroup();
						final Aspect asp = dd.getAspect();
						final short simulationVariant = dd.getSimulationVariant();
						CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), simulationVariant, false);
						logMsg += subscrAck(so, atg, asp, simulationVariant, din, sm);
					}

					// Hier sollte es keinen Fehler geben, denn der SettingsManager sollte nur gueltige DIDs liefern:
					archMgr.getDavCon().subscribeReceiver(
							inQueuesMgr.getArchiveDataReceiver(), did.getObject(), dd, ReceiveOptions.delayed(), ReceiverRole.receiver()
					);
					logMsg += Debug.NEWLINE + i + ".: Empfangsanmeldung fuer " + did.getObject() + ", " + dd + " erfolgreich";
					logger.finest(logMsg);
				}
				catch(ConfigurationException e) {
					sm.subscrFaults++;
					logger.error(i + ".: Empfangsanmeldung fuer " + did.getObject() + ", " + dd + " fehlgeschlagen: " + e.getMessage());
					successful = false;
				}
			}
			logger.info(
					"Anmeldungen fuer Simulationsvariante " + ard.getSimVar() + " abgeschlossen:" + Debug.NEWLINE + Util.leadBlank(i, 6)
					+ " Datenidentifikationen angemeldet davon " + Util.leadBlank(sm.subscrFaults, 6) + " Fehler" + Debug.NEWLINE + Util.leadBlank(i, 6)
					+ " Quittierungen angemeldet davon " + Util.leadBlank(sm.qSubscrFault, 6) + " Fehler"
			);
			ard.doneSubscribing(successful);
		}
	}

	private void processParams() {
		if(printParams) {
			if(paramDumpStream != null) closeParamDumpStream(paramDumpStream);
		}
		else {
			logger.info("EndOfSettingsListener: Archivparametrierung fr " + totalDIDs + " Datenidentifikationen erhalten. Starte An/Abmeldungen...");
			processSubscriptions();
			logger.info(subscrMarker.printStatistics());
			subscrMarker.reset();	// Speicher sparen...
			totalDIDs = 0;
		}
	}

	/** Beendet den ArchivConfig-Thread. Die Methode blockiert (aus Gruenden der Einfachheit), bis der Thread beendet ist. */
	public void waitForTermination() {
		isTerminated = false;
		shallTerminate = true;

		if(getState() != Thread.State.NEW) {
			interrupt();
			try {
				synchronized(terminationMutex) {
					while(!isTerminated) terminationMutex.wait();
				}
			}
			catch(InterruptedException e) {
				isTerminated = true;
			}
		}
	}

	/**
	 * Erzeugt einen Listener der benachrichtigt wird, wenn ein Parametrierungszyklus durchlaufen ist.
	 *
	 * @return Listener
	 */
	public EndOfParamListener createEndOfParamListener() {
		EndOfParamListener eopl = new EndOfParamListener();
		synchronized(endOfParamListeners) {
			endOfParamListeners.add(eopl);
		}
		return eopl;
	}

	/**
	 * Initialisiert den Thread und startet durch {@link #restartSubscribeArchiveParams()} die Uebernahme der Archivparametrierung.
	 *
	 * @param iqMgr Empfaenger der Datenanmeldung
	 * @param dTree Baum, in den die Datenidentifikationen zusammen mit der Parametrierung eingetragen werden
	 *
	 * @throws ConfigurationException
	 */
	public void startSubscribeArchiveParams(InQueuesMgr iqMgr, DataIdentTree dTree) throws ConfigurationException {
		printParams = false;
		inQueuesMgr = iqMgr;
		didTree = dTree;
		start();
		restartSubscribeArchiveParams();
	}

	/**
	 * Startet die Uebernahme der Archivparametrierung und meldet die zu archivierenden Datenidentifikation an. Diese Methode wird auch nach einem
	 * Verbindungsabbruch zum DAV aufgerufen.
	 *
	 * @throws ConfigurationException
	 */
	public void restartSubscribeArchiveParams() throws ConfigurationException {
		subscrMarker = new SubscrMarker();
		settingsMgr = createSettingsMgr();
		settingsMgr.start();
	}

	/** Beendet den SettingsManager und meldet die Parametrierung des Archivsystems ab. */
	public void unsubscribeArchiveParams() {
		if(settingsMgr != null) {
			settingsMgr.removeEndOfSettingsListener(eofSettingsLstnr);
			settingsMgr.removeUpdateListener(updateSettingsLstnr);
			settingsMgr.stop();
			settingsMgr = null;
		}
	}

	/**
	 * Meldet die zu archivierenden Daten der angegebenen Simulationsvariante ab.
	 *
	 * @param simVar Simulationsvariante. Falls < 0 werden alle zu archivierenden Daten abgemeldet
	 */
	private void unsubscribeArchiveData(int simVar) {
		logger.info("Start Abmeldung aller zu archivierenden Datenidentifikationen" + (simVar < 0 ? "" : (" der Simulationsvariante " + simVar)));
		int numDID = 0;
		if(didTree != null) {
			DataIdentTreeIterator it = didTree.createIterator();
			while(it.getNext()) {
				DataIdentNode din = it.getDataIdentNode();
				if((simVar < 0 || it.getSV() == simVar) && din.isArSParameterized() && din.arSParamIsArchivieren()) {
					try {
						SystemObject so = archMgr.getObj(it.getObjId());
						DataDescription dd = new DataDescription(archMgr.getAtg(it.getAtgId()), archMgr.getAsp(it.getAspId()), (short)it.getSV());
						archMgr.getDavCon().unsubscribeReceiver(inQueuesMgr.getArchiveDataReceiver(), so, dd);
						logger.finest("Abmeldung fuer " + so + ", " + dd + " erfolgreich");
						numDID++;
					}
					catch(ConfigurationException e) {
						logger.error(
								"Abmelden von " + Util.did2Str(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV()) + " fehlgeschlagen: " + e.getMessage()
						);
					}
				}
			}
		}
		logger.info(numDID + " zu archivierende Datenidentifikationen " + (simVar != -1 ? ("fuer SimVar==" + simVar) : "") + "abgemeldet");
	}

	/** Meldet alle zu archivierenden Daten ab */
	public void unsubscribeArchiveData() {
		unsubscribeArchiveData(-1);
	}

	/**
	 * Meldet alle zu einer Simulation gehoerenden Daten (= zu archivierende Daten + Quittungen) ab. Da es nicht auf einfache Weise moeglich ist, festzustellen,
	 * wann keine Datensaetze einer bestimmten SimVar mehr in der Archiv-Queue sind, kann man nicht bis dahin warten, bevor die Quittungen abgemeldet werden. Es
	 * kann also sein, dass einige Datensaetze der Simulation nicht mehr quittiert werden.
	 */
	public void unsubsrcibeSimVarData(int simVar) {
		unsubscribeArchiveData(simVar);
		unsubscribeArchiveDataAck(simVar);

		DataIdentTreeIterator it = didTree.createIterator();
		while(it.getNext()) if(it.getSV() == simVar) it.getDataIdentNode().arsParamSetParams(null);
	}


	/**
	 * Meldet die Quittierungsaspekte der genannte Simulationsvariante ab. Dies muss nach {@link #unsubscribeArchiveData()} geschehen, damit die Datensaetze,
	 * die noch in der Queue sind, nach der Archivierung quittiert werden koennen.
	 */
	private void unsubscribeArchiveDataAck(int simVar) {
		logger.info("Start Abmeldung aller Quittierungsaspekte" + (simVar < 0 ? "" : (" der Simulationsvariante " + simVar)));
		ClientDavInterface davCon = archMgr.getDavCon();
		int numQ = 0;
		if(didTree != null) {
			DataIdentTreeIterator it = didTree.createIterator();
			while(it.getNext()) {
				DataIdentNode din = it.getDataIdentNode();
				if((simVar < 0 || it.getSV() == simVar) && din.isArSParameterized() && din.arSParamIsQuittierenValid()) {
					try {
						SystemObject so = davCon.getDataModel().getObject(it.getObjId());
						AttributeGroup atg = (AttributeGroup)davCon.getDataModel().getObject(it.getAtgId());
						Aspect asp = (Aspect)davCon.getDataModel().getObject(din.arSParamGetQuittieren());
						CacheManager.getInstance().setCachingEnabled(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV(), true);
						davCon.unsubscribeSender(inQueuesMgr.getDataAckSender(), so, new DataDescription(atg, asp, (short)it.getSV()));
						numQ++;
					}
					catch(ConfigurationException e) {
						logger.error(
								"Abmelden der Quittierung " + Util.did2Str(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV()) + " fehlgeschlagen: "
								+ e.getMessage()
						);
					}
				}
			}
		}
		logger.info(numQ + " Quittierungsaspekte " + (simVar != -1 ? ("fuer SimVar==" + simVar) : "") + "abgemeldet");
	}

	/**
	 * Meldet alle Quittierungsaspekte ab. Dies muss nach {@link #unsubscribeArchiveData()} geschehen, damit die Datensaetze, die noch in der Queue sind, nach
	 * der Archivierung quittiert werden koennen.
	 */
	public void unsubscribeArchiveDataAck() {
		unsubscribeArchiveDataAck(-1);
	}


	/**
	 * Gibt die Parametrierung auf dem uebergebenen Stream aus.
	 *
	 * @param dumpFile Ausgabestream. Falls null, wird auf stdout ausgegeben
	 *
	 * @throws ConfigurationException
	 * @throws InterruptedException
	 */
	public void printArchiveParams(File dumpFile) throws Exception {
		printParams = true;
		paramDumpStream = dumpFile == null ? null : createParamDumpStream(dumpFile);
		noAnswer = false;
		gotFirstPrintParam = false;
		final EndOfParamListener eopl = createEndOfParamListener();

		Thread watchdog = new Thread() {
			public void run() {
				try {
					sleep(PARAM_TIMEOUT);
					if(!gotFirstPrintParam) {
						// Wenn schon angefangen wurde nicht unterbrechen
						noAnswer = true;
						eopl.gotParams();
					}
				}
				catch(InterruptedException e) {
				}
			}
		};
		watchdog.start();
		start();
		settingsMgr = createSettingsMgr();
		settingsMgr.start();
		eopl.waitForParams();
		watchdog.interrupt();
		if(noAnswer) throw new Exception("Kein Parameter-Datensatz nach " + (PARAM_TIMEOUT / 1000) + " Sekunden erhalten.");
	}

	private FileOutputStream createParamDumpStream(File dumpFile) {
		try {
			if(dumpFile.isDirectory()) dumpFile = new File(dumpFile, DEFAULT_PARAMS_FILE);
			if(dumpFile.exists()) archMgr.quitError("Ausgabedatei '" + dumpFile.getAbsolutePath() + "' fr die Archivparametrierung existiert bereits.");
			if(!dumpFile.createNewFile()) {
				archMgr.quitError("Ausgabedatei '" + dumpFile.getAbsolutePath() + "' fr die Archivparametrierung konnte nicht angelegt werden.");
			}
			if(!dumpFile.canWrite()) archMgr.quitError("Ausgabedatei '" + dumpFile.getAbsolutePath() + "' kann nicht geschrieben werden.");
			return new FileOutputStream(dumpFile);
		}
		catch(IOException e) {
			archMgr.quitError("Ausgabedatei '" + dumpFile.getAbsolutePath() + "' fr die Archivparametrierung konnte nicht angelegt werden: " + e.getMessage());
		}
		return null;
	}

	private void closeParamDumpStream(FileOutputStream ds) {
		try {
			ds.close();
		}
		catch(IOException e) {
			logger.error("Fehler beim Schreiben der Ausgabedatei fr die Archivparametrierung" + e.getMessage());
		}
	}

	private String unsubscrAck(SystemObject so, AttributeGroup atg, int sv, long qAspId) {
		subscrMarker.qUnsubscrTotal++;
		try {
			Aspect qAsp = archMgr.getAsp(qAspId);
			archMgr.getDavCon().unsubscribeSender(inQueuesMgr.getDataAckSender(), so, new DataDescription(atg, qAsp, (short)sv));
			return "Sendeabmeldung fuer Quittierungsaspekt '" + qAsp.getNameOrPidOrId() + "' erfolgreich.";
		}
		catch(IllegalArgumentException e) {
			// das sollte mit den gueltigen/ungueltigen Q-Asp vermieden werden:
			logger.error("Sendeabmeldung eines illegalen Quittierungsaspektes: " + e.getMessage());
			return "Sendeabmeldung eines illegalen Quittierungsaspektes: " + e.getMessage();
		}
		catch(ConfigurationException e) {
			subscrMarker.qUnsubscrFault++;
			return "Sendeabmeldung fuer Quittierungsaspekt 'id=" + qAspId + "' fehlgeschlagen: " + e.getMessage();
		}
	}

	/**
	 * Meldet die Quittungs an. Falls nicht erfolgreich, wird dies im DataIdentNode vermerkt (durch negative ID-1).
	 *
	 * @param so  Systemobjekt
	 * @param atg Attributgruppe
	 * @param sv  SimVar
	 * @param din DataIdentnode
	 *
	 * @return Log-Message
	 */
	private String subscrAck(SystemObject so, AttributeGroup atg, Aspect asp, int sv, DataIdentNode din, SubscrMarker sm) {
		Exception ex;
		long qAspId = din.arSParamGetQuittieren();
		if(sm != null) sm.qSubscrTotal++;
		try {
			if(asp.getId() == qAspId) throw new Exception("AspektID==QuittungsaspektID");
			Aspect qAsp = archMgr.getAsp(qAspId);
			archMgr.getDavCon().subscribeSender(
					archMgr.getInQueuesMgr().getDataAckSender(), so, new DataDescription(atg, qAsp, (short)sv), SenderRole.source()
			);
			din.arSParamMarkQuittierenInvalid(false);
			return "Sendeanmeldung fuer Quittierungsaspekt '" + qAsp.getNameOrPidOrId() + "' erfolgreich.";
		}
		catch(Exception e) {
			ex = e;
		}

		if(sm != null) sm.qSubscrFault++;
		din.arSParamMarkQuittierenInvalid(true);
		logger.fine("Sendeanmeldung fuer Quittierungsaspekt 'id=" + qAspId + "' fuer " + so + ", " + atg + "' fehlgeschlagen: " + ex.getMessage());
		return "Sendeanmeldung fuer Quittierungsaspekt 'id=" + qAspId + "' fehlgeschlagen: " + ex.getMessage();
	}

	/**
	 * Diese Methode arbeitet den {@link SubscrMarker} ab, der in den Callback-Methoden des {@link SettingsManager} gefuellt wurde. Es werden gemaess
	 * Parametrierung alle An- und Abmeldungen fuer die Datensaetze als auch fuer die Quittungen durchgefuehrt. Die {@link DataIdentNode}s wurden bereits mit der
	 * aktuellen Parametrierung besetzt.
	 */
	private void processSubscriptions() {
		SystemObject so;
		AttributeGroup atg;
		Aspect asp;
		DataDescription dd;
		DataIdentTreeIterator it = didTree.createIterator();
		ClientDavInterface davCon = archMgr.getDavCon();

		long i = 0, t = Util.startTimer();
		while(it.getNext()) {
			if(shallTerminate) return;

// Sollen Simulationen hier mit aufgenommen werden?
//			if (it.getSV() != 0)
//				continue;	

			DataIdentNode din = it.getDataIdentNode();
			SubscrInfo si = subscrMarker.getSubscrInfo(din);
			if(si != null) {
				// fuer diesen DataIdentNode hat sich im letzten Parametrierungslauf etwas geaendert
				i++;
				String logMsg = "";
				try {
					so = archMgr.getObj(it.getObjId());
					atg = archMgr.getAtg(it.getAtgId());
					asp = archMgr.getAsp(it.getAspId());
					dd = new DataDescription(atg, asp, (short)it.getSV());
				}
				catch(ConfigurationException e) {
					logger.warning(
							"An/Abmeldung von " + Util.did2Str(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV()) + " fehlgeschlagen: " + e.getMessage()
					);
					continue;
				}

				// Objekte erfolgreich erhalten, nun gibt es 6 Faelle, in denen An/Abmeldungen notwendig sein koennen:

				if(si.kind == SubscrInfo.NEW_PARAM_ARCH || si.kind == SubscrInfo.RE_PARAM_ARCH) {
					// *** Anmelden wegen Neuparametrierung (mit arch=ja) oder Umparametrierung (mit arch=ja) ***

					// laut TAnfArS 5.1.2.5 muessen bei Aenderung des Attributes "archivieren" offene Cont. abgeschlossen werden.
					if(si.kind == SubscrInfo.RE_PARAM_ARCH) inQueuesMgr.insertCloseContainer(so.getId(), new ArchiveTask.CloseContainerObject(dd));

					try {
						// Beim ersten DS nach Anmeldung darf der DIdx ausnahmsweise gleich dem letzten DIdx sein (wenn sich die DID nicht
						// geaendert hat). Um dies feststellen zu koennen, mussen folgende Flags gesetzt werden:
						din.setFirstDataAfterSubscription(ArchiveDataKind.ONLINE, true);
						din.setFirstDataAfterSubscription(ArchiveDataKind.ONLINE_DELAYED, true);

						// Hier sollte es keinen Fehler geben, denn der SettingsManager sollte nur gueltige DIDs liefern:
						davCon.subscribeReceiver(inQueuesMgr.getArchiveDataReceiver(), so, dd, ReceiveOptions.delayed(), ReceiverRole.receiver());
						logMsg += "\n" + i + ".: Empfangsanmeldung fuer " + so + ", " + dd + " erfolgreich";

						// wenn das erfolgreich war, Quittung anmelden:
						if(din.arSParamIsQuittieren()) {
							CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), it.getSV(), false);
							logMsg += subscrAck(so, atg, asp, it.getSV(), din, subscrMarker);       // setzt Quittung valid/invalid
						}
						else {
							din.arSParamMarkQuittierenInvalid(false);
						}

						if(si.kind == SubscrInfo.RE_PARAM_ARCH) logMsg += "; aktuelle Container wurden vor der Anmeldung geschlossen";
						logger.finest(logMsg);
					}
					catch(ConfigurationException e) {
						subscrMarker.subscrFaults++;
						logger.error(i + ".: Empfangsanmeldung fuer " + so + ", " + dd + " fehlgeschlagen: " + e.getMessage());
					}
					catch(IllegalArgumentException iae) {
						subscrMarker.subscrFaults++;
						logger.error(i + ".: Empfangsanmeldung fuer " + so + ", " + dd + " fehlgeschlagen: " + iae.getMessage());
					}
				}
				else if(si.kind == SubscrInfo.UN_PARAM || si.kind == SubscrInfo.RE_PARAM_NOARCH) {
					// *** Abmelden (erst DID, wenn erfolgreich die Quittung (ist die konsistentere Loesung))
					try {
						davCon.unsubscribeReceiver(inQueuesMgr.getArchiveDataReceiver(), so, dd);
						logMsg += i + ".: Empfangsabmeldung fuer " + so + ", " + dd + " erfolgreich";

						// laut TAnfArS 5.1.2.5 muessen bei Aenderung des Attributes "archivieren" oder Abmeldung offene Cont. abgeschlossen werden.
						inQueuesMgr.insertCloseContainer(so.getId(), new ArchiveTask.CloseContainerObject(dd));
						logMsg += "; aktuelle Container werden geschlossen";

						// Abmeldung der Quittungsaspekte an dieser Stelle:
						// Dies bedeutet, dass bei DID, die vorher mit Quittierung
						// parametriert waren, und von denen noch welche in der Queue sind, keine Quittungen
						// mehr gesendet werden koennen. Siehe ArchivTask.sendReceipt()

						if(si.kind == SubscrInfo.UN_PARAM && din.arSParamIsQuittierenValid()) {
							CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), it.getSV(), true);
							final long qAsp = din.arSParamGetQuittieren();
							logMsg += "; " + unsubscrAck(so, atg, it.getSV(), qAsp);
						}
						else if(si.kind == SubscrInfo.RE_PARAM_NOARCH && si.hasOldAspect()) {
							CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), it.getSV(), true);
							final long qAsp = si.oldQAspectId;
							logMsg += "; " + unsubscrAck(so, atg, it.getSV(), qAsp);
						}

						// wenn nicht archiviert wird, kann auch nicht quittiert werden
						din.arSParamMarkQuittierenInvalid(true);

						if(si.kind == SubscrInfo.UN_PARAM) din.arsParamSetParams(null);	// nicht mehr Teil der aktuellen Parametrierung

						logger.finest(logMsg);
					}
					catch(ConfigurationException e) {
						subscrMarker.unsubscrFaults++;
						logger.error(i + ".: Empfangsabmeldung fuer " + so + ", " + dd + " fehlgeschlagen: " + e.getMessage());
					}
				}
				else if(si.kind == SubscrInfo.RE_PARAM_PARAM) {
					// *** Ggf. Abmelden des alten und Anmelden des neuen Q-Aspekts:
					CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), it.getSV(), true);
					if(si.hasOldAspect()) logMsg += unsubscrAck(so, atg, it.getSV(), si.oldQAspectId);

					if(din.arSParamIsArchivieren()) din.arSParamMarkQuittierenInvalid(false);
					if(din.arSParamIsArchivieren() && din.arSParamIsQuittieren()) {
						CacheManager.getInstance().setCachingEnabled(so.getId(), atg.getId(), asp.getId(), it.getSV(), false);
						logMsg += "; " + subscrAck(so, atg, asp, it.getSV(), din, subscrMarker);
					}

					if(si.closeContainer) {
						// laut TAnfArS 5.1.2.5 muessen bei Aenderung der Attributes "sichern" oder "vorhalten" offene Cont. abgeschlossen werden.
						inQueuesMgr.insertCloseContainer(so.getId(), new ArchiveTask.CloseContainerObject(dd));
						logMsg += "; aktuelle Container werden geschlossen";
					}
					logger.finest(logMsg);
				}
				else if(si.kind == SubscrInfo.NEW_PARAM_NOARCH) {
					// wenn nicht archiviert wird, kann auch nicht quittiert werden
					din.arSParamMarkQuittierenInvalid(true);
				}
			}
			else	// keine aktuelle Parametrierung fuer diesen DIN vorhanden
			{
				din.arsParamSetParams(null);	// nach DAV-Absturz notwendig, wenn nicht abgemeldet wurde
			}
		}
		logger.fine(i + " An/Abmeldungen durchgefhrt in " + Util.stopTimer(t));
	}

	/**
	 * Vermerkt einen DataIdentNode als reparametriert.
	 *
	 * @param did  DataIdentNode
	 * @param newD Neue Parameterdaten
	 */
	private void prepareReparam(DataIdentification did, Data newD) {
		try {
			DataIdentNode din = didTree.getDataIdentNode(did);
			boolean archivieren = getBool(newD, K.ARCHIVIEREN);
			boolean closeContainer =
					(din.arSParamGetVorhalten() != newD.getTimeValue(K.VORHALTEN).getSeconds()) || (din.arSParamIsSichern() != getBool(newD, K.SICHERN));

			if(!din.arSParamIsArchivieren() && archivieren) {
				subscrMarker.reParamArch(din);		// Ab jetzt archivieren (Anmeldung vermerken)
			}
			else if(din.arSParamIsArchivieren() && !archivieren) {
				// Ab jetzt nicht mehr archivieren (Abmeldung u. ggf. alten Q-Aspekt vermerken)
				if(din.arSParamIsQuittierenValid()) {
					subscrMarker.reParamNoArch(din, din.arSParamGetQuittieren());
				}
				else {
					subscrMarker.reParamNoArch(din);
				}
			}
			else {
				// Sonstige nderung der Parametrierung (z.B. der Quittierung)
				if(din.arSParamIsQuittierenValid() && din.arSParamIsQuittieren()) {
					subscrMarker.reParam(din, din.arSParamGetQuittieren(), closeContainer); // alte Quittierung zum Abmelden vermerken
				}
				else {
					subscrMarker.reParam(din, closeContainer);
				}
			}
			din.arsParamSetParams(newD);				// Parametrierung berschreiben (deshalb alten Quittungsaspekt merken)
		}
		catch(DIdNodeNotFoundException e) {
			logger.warning("Umparametrierung: Datenidentifikation nicht im DataIdentTree vorhanden (Umparametrierung wird nicht durchgefhrt): " + did);
		}
	}

	/**
	 * Vermerkt einen DataIdentNode als neu parametriert. Der Node wird erzeugt, falls er nicht existiert. Dies passiert in der Callback-Methode. Die
	 * DataIdentNodes sind also immer vor {@link #processParams()} angelegt. D.h. sobald dort das erste Datum angemeldet wird, kann es archiviert werden.
	 *
	 * @param did  Datenidentifikation
	 * @param newD Neue Parameterdaten
	 */
	private void prepareNewParam(DataIdentification did, Data newD) {
		DataIdentNode din;
		boolean notCreated = false;
		try {
			din = didTree.getDataIdentNode(did);
		}
		catch(DIdNodeNotFoundException e) {
			din = new DataIdentNode();
			notCreated = true;
		}
		din.arsParamSetParams(newD);				// Parametrierung vermerken
		if(notCreated) didTree.putDataIdentNode(did, din);
		if(din.arSParamIsArchivieren()) {
			subscrMarker.newParamArch(din);		// Als zu archivieren vermerken
		}
		else {
			subscrMarker.newParamNoArch(din);	// Als nicht zu archiveren zu vermerken
		}
	}

	/**
	 * Vermerkt einen DataIdentNode als nicht mehr parametriert.
	 *
	 * @param did Datenidentifikation
	 */
	private void prepareUnparam(DataIdentification did) {
		try {
			subscrMarker.unParam(didTree.getDataIdentNode(did));	// Parametrierung wird spaeter auf null gesetzt
			// (wir brauchen ggf. den alten Q-Aspekt)
		}
		catch(DIdNodeNotFoundException e) {
			logger.warning("Entparametrierung: Datenidentifikation nicht im DataIdentTree vorhanden (Entparametrierung wird nicht durchgefhrt): " + did);
		}
	}

	/**
	 * Schreibt die empfangene Parametrierung auf den bergebenen Ausgabestrom oder auf stdout.
	 *
	 * @param dumpStream    Ausgabestrom. Falls null, wird auf stdout geschrieben.
	 * @param separatorLine Zeile, die zwischen Datenstzen eingefgt wird (\n wird angehngt).
	 * @param did           Datenidentifikation
	 * @param data          Parametrierung
	 */
	private void printParam(OutputStream dumpStream, String separatorLine, DataIdentification did, Data data) {
		gotFirstPrintParam = true;
		StringBuilder sb = new StringBuilder();
		DataDescription dd = did.getDataDescription();
		sb.append(
				separatorLine + "\nobj   " + did.getObject().getPid() + " (id=" + did.getObject().getId() + ") '" + did.getObject().getName() + "'\natg   "
				+ dd.getAttributeGroup().getPid() + " (" + dd.getAttributeGroup().getId() + ") '" + dd.getAttributeGroup().getName() + "'\nasp   "
		);
		sb.append(dd.getAspect().getPid() + " (" + dd.getAspect().getId() + ") '" + dd.getAspect().getName() + "'\nsv    " + dd.getSimulationVariant() + "\n");
		if(data != null) {
			sb.append("einst archiv.: " + (getBool(data, K.ARCHIVIEREN) ? "ja" : "nein"));
			sb.append(" / sich.: " + (getBool(data, K.SICHERN) ? "ja" : "nein"));
			sb.append(
					" / quitt.: " + (data.getArray(K.QUITTIEREN).getLength() > 0
					                 ? data.getArray(K.QUITTIEREN).getItem(0).asReferenceValue().getValueText()
					                 : "---")
			);
			sb.append(" / vorh.: " + data.getTimeValue(K.VORHALTEN).getSeconds());
			int nachf = data.getArray(K.NACHFORDERN).getLength();
			if(nachf == 0) {
				sb.append(" / nachf.: ---");
			}
			else {
				sb.append("\n      nachf.:\n");
				for(int j = 0; j < nachf; j++) {
					sb.append("      " + data.getArray(K.NACHFORDERN).getItem(j).asTextValue().getValueText());
				}
			}
			sb.append("\n");
		}
		else {
			sb.append("   ## Fehler: Data==null");
		}

		if(dumpStream != null) {
			try {
				dumpStream.write(sb.toString().getBytes());
				dumpStream.flush();
			}
			catch(IOException e) {
				logger.error("Fehler beim Schreiben der Ausgabedatei fr die Archivparametrierung: " + e.getMessage());
			}
		}
		else {
			System.out.print(sb.toString());
		}
	}

	private boolean getBool(Data data, String key) {
		return data.getUnscaledValue(key).intValue() > 0;
	}

	/**
	 * Mit dieser Klasse wird whrend der Parameterbernahme in den Callback-Methoden vermerkt fr welche Datenidentifikation (= {@link DataIdentNode}) was
	 * parametriert wurde. Dadurch ist eine Trennung von Parameterbernahme und An-/Abmeldung mglich. Datenidentifikationen knnen dadurch falls notwendig fr
	 * blockweise An-/Abmeldung zusammengefasst werden. Die Parametrierung, die im {@link DataIdentNode} vermerkt ist, ist allein nicht ausreichend, da auch die
	 * Parameteraenderung relevant ist. Gespeichert wird dies durch eine Map von {@link DataIdentNode} nach {@link SubscrInfo}. Da die neuen Parameter sofort im
	 * {@link DataIdentNode} gespeichert werden, ist es in einigen Faellen erforderlich, den alten Quittierungsaspket hier mitzuspeichern.
	 * <p/>
	 * Unterschieden wird nach:
	 * <p/>
	 * -- Neu parametriert, zu archivieren -- Neu parametriert, nicht zu archivieren -- Nicht mehr parametriert -- Umparametriert, zu archivieren --
	 * Umparametriert, nicht zu archivieren (der ehemalige Quittierungsaspekt wird gespeichert, falls vorhanden) -- Umparametriert, Archivierung nicht gendert,
	 * jedoch ggf. Quittierungsaspekt
	 *
	 * @author AlexanderS
	 */
	public static class SubscrMarker {

		// Platz sparen beim Aufbau der Markierungen, daher DAG-Struktur soweit wie moeglich:
		public static SubscrInfo SI_NEW_PARAM_ARCH = new SubscrInfo(SubscrInfo.NEW_PARAM_ARCH), SI_NEW_PARAM_NOARCH = new SubscrInfo(SubscrInfo.NEW_PARAM_NOARCH), SI_UN_PARAM = new SubscrInfo(
				SubscrInfo.UN_PARAM
		), SI_RE_PARAM_ARCH = new SubscrInfo(SubscrInfo.RE_PARAM_ARCH), SI_RE_PARAM_NOARCH = new SubscrInfo(SubscrInfo.RE_PARAM_NOARCH), SI_RE_PARAM_PARAM = new SubscrInfo(
				SubscrInfo.RE_PARAM_PARAM
		), SI_RE_PARAM_PARAM_CLOSE = new SubscrInfo(SubscrInfo.RE_PARAM_PARAM);

		static {
			SI_RE_PARAM_PARAM_CLOSE.closeContainer = true;
		}

		public static class SubscrInfo {

			public static int NEW_PARAM_ARCH = 1, NEW_PARAM_NOARCH = 2, UN_PARAM = 3, RE_PARAM_ARCH = 4, RE_PARAM_NOARCH = 5, RE_PARAM_PARAM = 6;

			/** Was ist geschehen (Neu-/Ent-/Umparametrierung) */
			public int kind = -1;

			/** In einigen Faellen braucht man den alten Quittungsaspekt */
			public long oldQAspectId = -1;

			public boolean closeContainer = false;

			public SubscrInfo(int kind) {
				this.kind = kind;
			}

			public SubscrInfo(int kind, long oldQAspectId) {
				this.kind = kind;
				this.oldQAspectId = oldQAspectId;
			}

			public boolean hasOldAspect() {
				return oldQAspectId != -1;
			}

			public SubscrInfo clone() {
				SubscrInfo clone = new SubscrInfo(kind, oldQAspectId);
				clone.closeContainer = closeContainer;
				return clone;
			}
		}

		private Map<DataIdentNode, SubscrInfo> marker = new HashMap<DataIdentNode, SubscrInfo>();

		// An- und Abmeldefehler mitzaehlen
		public int subscrFaults = 0, unsubscrFaults = 0, qSubscrTotal = 0, qUnsubscrTotal = 0, qUnsubscrFault = 0, qSubscrFault = 0;

		/**
		 * Vermerken, dass dieser DataIdentNode neu parametriert wurde und archiviert wird. Anmeldung der DID und ggf. der Quittierung erforderlich.
		 *
		 * @param din
		 */
		public void newParamArch(DataIdentNode din) {
			marker.put(din, SI_NEW_PARAM_ARCH);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode neu parametriert wurde und nicht archiviert wird. Keine An-/Abmeldungen notwendig, nur zur Statistik vermerken.
		 *
		 * @param din
		 */
		public void newParamNoArch(DataIdentNode din) {
			marker.put(din, SI_NEW_PARAM_NOARCH);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode nicht mehr parametriert ist. Abmeldung der DID und ggf. der alten (noch im din enthaltenen) Quittierung erforderlich.
		 *
		 * @param din
		 */
		public void unParam(DataIdentNode din) {
			marker.put(din, SI_UN_PARAM);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode auf archivieren umparametriert wurde. Anmeldung der DID und ggf. der Quittierung erforderlich.
		 *
		 * @param din
		 */
		public void reParamArch(DataIdentNode din) {
			marker.put(din, SI_RE_PARAM_ARCH);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode auf nicht archivieren umparametriert wurde. Abmeldung der DID und der alten Quittierung erforderlich. Da die
		 * Quittierung im DataIdentNode ueberschrieben worden sein kann, muss sie mitgespeichert werden.
		 *
		 * @param din
		 */
		public void reParamNoArch(DataIdentNode din, long oldQAspectId) {
			SubscrInfo sc = SI_RE_PARAM_NOARCH.clone();
			sc.oldQAspectId = oldQAspectId;
			marker.put(din, sc);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode auf nicht archivieren umparametriert wurde. Abmeldung der DID erforderlich, keine Quittierung abzumelden.
		 *
		 * @param din
		 */
		public void reParamNoArch(DataIdentNode din) {
			marker.put(din, SI_RE_PARAM_NOARCH);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode umparametriert wurde, was keinen Einfluss auf die Archivierung hat. Abmeldung der Quittierung erforderlich. Wenn sich
		 * bestimmte Werte der Parametrierung geaendert haben, mussen aktuelle Container abgeschlossen werden.
		 *
		 * @param din
		 */
		public void reParam(DataIdentNode din, long oldQAspectId, boolean closeCont) {
			SubscrInfo sc = SI_RE_PARAM_PARAM.clone();
			sc.oldQAspectId = oldQAspectId;
			sc.closeContainer = closeCont;
			marker.put(din, sc);
		}

		/**
		 * Vermerken, dass dieser DataIdentNode umparametriert wurde, jedoch nicht in seiner Quittierungseigenschaft. Wenn sich bestimmte Werte der Parametrierung
		 * geaendert haben, mussen aktuelle Container abgeschlossen werden.
		 *
		 * @param din
		 */
		public void reParam(DataIdentNode din, boolean closeCont) {
			marker.put(din, closeCont ? SI_RE_PARAM_PARAM_CLOSE : SI_RE_PARAM_PARAM);
		}

		/**
		 * Liefert die vorgemerkte Markierung.
		 *
		 * @param din
		 */
		public SubscrInfo getSubscrInfo(DataIdentNode din) {
			return marker.get(din);
		}

		/** Lschen der gespeicherten Markierungen. */
		public void reset() {
			subscrFaults = 0;
			unsubscrFaults = 0;
			qSubscrTotal = 0;
			qUnsubscrTotal = 0;
			qUnsubscrFault = 0;
			qSubscrFault = 0;
			marker.clear();
		}

		/** Ausgabe der gespeicherten Markierungen. */
		private String printStatistics() {
			int newparam_arch = 0, newparam_noarch = 0, reparam_param = 0, reparam_subscr = 0, reparam_unsubscr = 0, unparam = 0, closeCont = 0;
			for(SubscrInfo si : marker.values()) {
				if(si.kind == SubscrInfo.NEW_PARAM_ARCH) {
					newparam_arch++;
				}
				else if(si.kind == SubscrInfo.NEW_PARAM_NOARCH) {
					newparam_noarch++;
				}
				else if(si.kind == SubscrInfo.UN_PARAM) {
					unparam++;
					closeCont++;
				}
				else if(si.kind == SubscrInfo.RE_PARAM_ARCH) {
					reparam_subscr++;
					closeCont++;
				}
				else if(si.kind == SubscrInfo.RE_PARAM_NOARCH) {
					reparam_unsubscr++;
					closeCont++;
				}
				else if(si.kind == SubscrInfo.RE_PARAM_PARAM) {
					reparam_param++;
					if(si.closeContainer) closeCont++;
				}
			}
			return "An/Abmeldungen abgeschlossen:\n  " + Util.leadBlank(newparam_arch + newparam_noarch, 6) + " Zugaenge (" + newparam_arch + " arch / "
			       + newparam_noarch + " noarch), " + unparam + " Abgaenge, " + (reparam_param + reparam_unsubscr + reparam_subscr) + " Aenderungen ("
			       + reparam_subscr + "->arch / " + reparam_unsubscr + "->noarch / " + reparam_param + " andere)" + ", insgesamt " + marker.size() + "\n  "
			       + Util.leadBlank(closeCont, 6) + " Datenidentifikationen abzuschliessen" + "\n  " + Util.leadBlank(subscrFaults, 6)
			       + " Fehler bei Anmeldungen von Datenidentifikationen" + "\n  " + Util.leadBlank(unsubscrFaults, 6)
			       + " Fehler bei Abmeldungen von Datenidentifikationen" + "\n  " + Util.leadBlank(qSubscrTotal, 6) + " Anmeldungen von Quittierungen, davon "
			       + qSubscrFault + " Fehler" + "\n  " + Util.leadBlank(qUnsubscrTotal, 6) + " Abmeldungen von Quittierungen, davon " + qUnsubscrFault
			       + " Fehler";
		}
	}

	/**
	 * Spezielle Klasse, mit der man auf das Ende der bernahme der nchsten Archiveinstellungen warten kann. Dies ist u.a. fuer Testzwecke hilfreich. Listener
	 * dieser Art koennen nur einmal verwendet werden, weil die Liste der Listener in der Klasse {@link ArchivConfig} nach jedem Zyklus geleert wird.
	 */
	public class EndOfParamListener {

		public boolean gotParams = false;

		private synchronized void gotParams() {
			gotParams = true;
			notifyAll();
		}

		public synchronized void waitForParams() throws InterruptedException {
			while(!gotParams) wait();
		}

		public String describeAwaitedObjects() {
			return "    " + ARS_PARAM_ATG_PID + "/" + ARS_PARAM_ASP_PID;
		}
	}
}
