/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt;

import de.bsvrz.ars.ars.mgmt.commands.ArSCmdInterface;
import de.bsvrz.ars.ars.mgmt.datatree.DataIdentTree;
import de.bsvrz.ars.ars.mgmt.simulation.SimulationManager;
import de.bsvrz.ars.ars.mgmt.tasks.BackupTask;
import de.bsvrz.ars.ars.mgmt.tasks.scheduler.TaskScheduler;
import de.bsvrz.ars.ars.persistence.PersistenceManager;
import de.bsvrz.ars.ars.persistence.CacheManager;
import de.bsvrz.dav.daf.main.CommunicationError;
import de.bsvrz.dav.daf.main.ConnectionException;
import de.bsvrz.dav.daf.main.InconsistentLoginException;
import de.bsvrz.dav.daf.main.MissingParameterException;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.DAVAppBase;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;

import java.io.File;
import java.util.*;

/**
 * Zentrale Archiv-Verwaltung.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @author Alexander Schmidt
 * @version $Revision: 10735 $ / $Date: 2012-11-19 18:14:45 +0100 (Mon, 19 Nov 2012) $ / ($Author: rs $)
 */
public class ArchiveManager extends DAVAppBase {

	private static long SYS_TIME_OFFSET = 0;			// nur fuer testzwecke

	public static long DATAINDEX_OFFSET = 0;			// nur fuer testzwecke

	/**
	 * Nur fuer Testzwecke; falls wahr, werden nur Datensaetze mit Daten archiviert. Es gibt dann keine Datensaetze folgender Art: "keine Quelle", "keine Daten",
	 * "pot. Datenluecke". Manche Testverfahren werden einfacher, wenn ausschliesslich Datensaetze archiviert werden, die auch gesendet wurden.
	 */
	public static boolean ONLY_DATA = false;

	private static final int DEFAULT_CMD_PORT = 4242;

	private static final String P_PRINT_PARAMS = "-nurParametrierungAusgeben";

	private static final String P_PRINT_PARAMS_FILE = "-parametrierungAusgabeDatei";

	private static final String P_PERS_PATH = "-persistenzVerzeichnis";

	private static final String P_CHECK_PERS_DIR = "-persistenzVerzeichnisPruefen";

	private static final String P_REMOTE_CONTROL = "-kommandoPort";

	private static final String P_BACKUP = "-backup";

	public static final String P_CHECK_PERS_DIR_DIR = "Verzeichnis";

	public static final String P_CHECK_PERS_DIR_HDR = "Header";

	public static final String P_CHECK_PERS_DIR_ALL = "Alles";

	public static final String P_REMOTE_CONTROL_NORC = "deaktiviert";

	public static final String P_REMOTE_CONTROL_DEFAULT = "default";

	private PersistenceManager persistenceMgr;

	private DataIdentTree dataIdentTree;

	private InQueuesMgr inQueuesMgr;

	private ArchivConfig archivConfig;

	private TaskScheduler taskScheduler;

	private SimulationManager simulationMgr = null;

	// Aufrufparameter:
	private boolean printParamsAndExit;

	private File printParamsFile;

	private String checkPersDirMode, archivePath, backupProps;

	private int cmdPortNum = -1;

	private boolean terminated = false;

	private boolean runningProperly = false;

	private Object _terminatedLock = new Object();

	/** Enthlt die Anzahl aktiver Archivanfragen je Applikation. Es sind nur fr die Applikationen Eintrge in der Map enthalten fr die mindestens eine aktive Archivanfrage in der Verarbeitung ist. */
	private TreeMap<SystemObject, Integer> _applicationToArchiveQueryCount;


	public ArchiveManager(String[] args) throws Exception {
		super(args, "Archivsystem", new DisconnectHandler() {
			// Initialer DisconnectHandler wird spter in der run()-Methode ersetzt
			public void handleDisconnect() throws MissingParameterException, InconsistentLoginException, CommunicationError, ConnectionException {
				Debug.getLogger().error("Verbindung zum Datenverteiler wurde frhzeitig terminiert, Archivsystem wird beendet");
				System.exit(1);
			}
		});
		_applicationToArchiveQueryCount = new TreeMap<SystemObject, Integer>();
		logger = Debug.getLogger();
		parseArguments(argList);
		dataIdentTree = new DataIdentTree();
		if(!printParamsAndExit) persistenceMgr = new PersistenceManager(this, archivePath);
		archivConfig = new ArchivConfig(this);
	}

	public boolean wasTerminated() {
		synchronized(_terminatedLock) {
			return terminated;
		}
	}

	/** @return aktuelle Systemzeit (die zu Testzwecken mit einem Offset versehen werden kann) */
	public static long archiveTime() {
		return System.currentTimeMillis() + SYS_TIME_OFFSET;
	}

	/**
	 * Verstellt die Systemzeit um offset Sekunden. NUR FUER TESTZWECKE!!!
	 *
	 * @param offsetSeconds Spanne in Sekunden
	 */
	public static void setSystemTimeOffset(long offsetSeconds) {
		SYS_TIME_OFFSET = offsetSeconds * 1000;
	}


	private void parseArguments(ArgumentList argLst) throws Exception {
		// Kommandozeilenparameter:
		CacheManager.getInstance().init(argList);
		printParamsAndExit = argLst.hasArgument(P_PRINT_PARAMS);
		printParamsFile = argLst.hasArgument(P_PRINT_PARAMS_FILE) ? argLst.fetchArgument(P_PRINT_PARAMS_FILE).asFile() : null;
		checkPersDirMode = argLst.hasArgument(P_CHECK_PERS_DIR) ? argLst.fetchArgument(P_CHECK_PERS_DIR).asString() : "";
		archivePath = argLst.hasArgument(P_PERS_PATH) ? argLst.fetchArgument(P_PERS_PATH).asString() : null;
		String cmdPort = argLst.hasArgument(P_REMOTE_CONTROL) ? argLst.fetchArgument(P_REMOTE_CONTROL).asString() : null;
		backupProps = argLst.hasArgument(P_BACKUP) ? argLst.fetchArgument(P_BACKUP).asString() : null;

		// Konsistenzpruefung Kommandozeilenparameter:
		if(checkPersDirMode.length() > 0 && archivePath == null) quitError("Ungltige Parameterkombination: " + P_CHECK_PERS_DIR + " erfordert " + P_PERS_PATH);

		if(checkPersDirMode.length() > 0 && printParamsAndExit) {
			quitError("Ungltige Parameterkombination: " + P_CHECK_PERS_DIR + " und " + P_PRINT_PARAMS + " nicht gleichzeitig mglich");
		}

		if(checkPersDirMode.length() > 0 && !(checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_DIR) || checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_HDR)
		                                      || checkPersDirMode.equalsIgnoreCase(P_CHECK_PERS_DIR_ALL))) {
			quitError(
					"Ungltiger Parameterwert: Fr " + P_CHECK_PERS_DIR + " sind folgende Werte mglich: [" + P_CHECK_PERS_DIR_DIR + ", " + P_CHECK_PERS_DIR_HDR
					+ ", " + P_CHECK_PERS_DIR_ALL + "]"
			);
		}

		if(!printParamsAndExit && printParamsFile != null) {
			quitError("Ungltige Parameterkombination: " + P_PRINT_PARAMS_FILE + " nur in Verbindung mit " + P_PRINT_PARAMS + " mglich");
		}

		if((!printParamsAndExit || checkPersDirMode.length() > 0) && archivePath == null) {
			quitError("Ungltige Parameterkombination: " + P_PERS_PATH + " nicht vorhanden");
		}

		if(cmdPort != null && !cmdPort.equalsIgnoreCase(P_REMOTE_CONTROL_NORC)) {
			if(cmdPort.equals(P_REMOTE_CONTROL_DEFAULT)) {
				cmdPortNum = DEFAULT_CMD_PORT;
			}
			else {
				try {
					cmdPortNum = Integer.parseInt(cmdPort);
				}
				catch(NumberFormatException e) {
					quitError("Ungueltiger Wert fuer " + P_REMOTE_CONTROL + ": '" + cmdPort + "'");
				}
			}
		}

		if(!(checkPersDirMode.length() > 0 || printParamsAndExit)) {
			// Wenn nur das PV geprueft werden oder die Parametrierung ausgegeben werden soll, brauchen wir kein Backup-Modul
			try {
				BackupTask.setBackupModul(backupProps);
			}
			catch(Exception e) {
				quitError(e.getMessage());
			}
		}
	}

	public void run() throws Exception {
		if(wasTerminated()) {
			logger.error("Das Archivsystem wurde bereits beendet.");
			throw new Exception("Das Archivsystem wurde bereits beendet.");
		}
		if(executeSpecialModeNoDAV()) {
			quit();
		}
		else {

			if(executeSpecialModeWithDAV())	// startet den DAV
			{
				quit();
			}
			else {
				// ##########################################################################################
				// hier gehts richtig los:
				logger.info("Archivsystem im Normalmodus gestartet");

				// Persistenzverzeichnis berprfen (evtl anderes ArS angemeldet?)
				String msg = persistenceMgr.createLockFile();
				if(msg != null) {
					quitError(msg);
				}
				else {
					setDisconnectHandler(new DisconnectHandler() {
						public void handleDisconnect() throws MissingParameterException, InconsistentLoginException, CommunicationError, ConnectionException {
							quitError("Verbindung zum Datenverteiler wurde terminiert, Archivsystem wird kontrolliert beendet");
						}
					});
					Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
						public void run() {
							quitErrorNoExit("Terminierungssignal empfangen, Archivsystem wird kontrolliert beendet");
						}
					}));
					// Keine Sperrdatei vorhanden, eine angelegt, wir koennen starten
					runningProperly = true;

					inQueuesMgr = new InQueuesMgr(this, dataIdentTree);
					taskScheduler = new TaskScheduler(this);
					simulationMgr = new SimulationManager(this);

					// Als erstes das Kommando-Interface starten
					if(cmdPortNum != -1) new ArSCmdInterface(this, cmdPortNum).start();

					// Loesch-/Nachfordern-/Sichern-Uhren starten:
					if(!wasTerminated()) taskScheduler.start();

					// PersistenceManager ist der einzige der einen Startup-Check braucht,
					// alle anderen Module haben keine persistenten Daten.
					if(!persistenceMgr.startupProcedure()) {
						quitError("Fehler im Persistenzverzeichnis. Archivsystem kann nicht gestartet werden.");
					}
					else {
						connectToDavQuit();	// Bis hierher sollte alles ohne DAV gehen

						if(!wasTerminated()) inQueuesMgr.getObjectsFromDav();	// jeder Task darf benoetigte Objekte holen
						if(!wasTerminated()) inQueuesMgr.startAllTasks();		// startet alle Tasks (mit Ausnahme von ArchivConfig)
						if(!wasTerminated()) inQueuesMgr.subscribeObjects();	// auf alle Parameter, Anfragen, etc. anmelden

						// mit der Anmeldung der zu archivierenden Daten gehts los...
						if(!wasTerminated()) archivConfig.startSubscribeArchiveParams(inQueuesMgr, dataIdentTree);

						if(!wasTerminated()) simulationMgr.start();				//Simulationsverwaltung starten
					}
				}
			}
		}
	}

	public void connectToDavQuit() {
		try {
			connectToDav();
		}
		catch(Exception e) {
			quitError("Es konnte keine Verbindung zum Datenverteiler aufgebaut werden.\n" + e.getMessage());
		}
	}

	protected void reconnectHandler() {
		super.reconnectHandler();
		if(!wasTerminated()) {
			try {
				if(!wasTerminated()) inQueuesMgr.subscribeObjects();	// auf alle Parameter, Anfragen, etc. anmelden
				if(!wasTerminated()) archivConfig.restartSubscribeArchiveParams();
			}
			catch(ConfigurationException e) {
				logger.error("Neuanmeldung nach Wiederherstellen der DAV-Verbindung fehlgeschlagen: " + e.getMessage());
			}
		}
	}


	private boolean executeSpecialModeNoDAV() {
		if(checkPersDirMode.length() > 0) {
			logger.info("Archivsystem zur Prfung des Persistenzverzeichnisses gestartet");
			persistenceMgr.checkPersistenceDir(checkPersDirMode);
			return true;
		}
		return false;
	}

	private boolean executeSpecialModeWithDAV() {
		if(printParamsAndExit) {
			connectToDavQuit();
			logger.info(
					"Archivsystem zur Parameterausgabe " + (printParamsFile == null ? "auf der Konsole" : ("in '" + printParamsFile.getAbsolutePath() + "'"))
					+ " gestartet"
			);
			try {
				archivConfig.printArchiveParams(printParamsFile);
			}
			catch(Exception e) {
				logger.error("Parameterausgabe fehlgeschlagen: " + e.getMessage());
			}
			return true;
		}
		return false;
	}

	/** Beendet alle Tasks, meldet App vom DAV ab. */
	private void cleanUp() {
		final boolean wasTerminated;
		synchronized(_terminatedLock) {
			wasTerminated = terminated;
			terminated = true;
		}
		if(wasTerminated) return;
		try {
			// Alle periodischen Tasks stoppen:
			logger.info("Beende Task Scheduler...");
			if(taskScheduler != null) taskScheduler.terminate();
			logger.info("Task Scheduler beendet");

			// Archivdaten abmelden
			logger.info("Melde Archivdaten ab...");
			if(archivConfig != null) archivConfig.unsubscribeArchiveData();
			logger.info("Archivdaten abgemeldet");

			// Archivparametrierung abmelden
			logger.info("Melde Archivparametrierung ab...");
			if(archivConfig != null) archivConfig.unsubscribeArchiveParams();
			logger.info("Archivparametrierung abgemeldet");

			//Simulationsverwaltung herunterfahren
			logger.info("Melde Simulation ab...");
			if(simulationMgr != null) simulationMgr.terminateTask();
			logger.info("Simulation abgemeldet");

			// Beenden aller (u.a. der ArchiveTasks) Tasks
			logger.info("Beende Tasks...");
			if(inQueuesMgr != null) inQueuesMgr.stopAllTasks();
			logger.info("Tasks beendet");

			// Alle anderen abonnierten Objekte abmelden (nach dem Beenden der ArchivQueryTasks, weil nach dem
			// abgebrochenem take() das Totengloeckchen ueber den DAV gesendet werden muss):
			logger.info("Melde sonstige Parameter ab...");
			if(getDavCon() != null) ConnectionManager.unsubscribeAll(getDavCon());
			logger.info("Sonstige Parameter abgemeldet");

			// Quittungen erst nach Leeren der Archivwarteschlange abmelden
			logger.info("Melde Quittungen ab...");
			if(archivConfig != null) archivConfig.unsubscribeArchiveDataAck();
			logger.info("Quittungen abgemeldet");

			// Warten bis SettingsManager fertig
			logger.info("Warte auf SettingsManager...");
			if(archivConfig != null) archivConfig.waitForTermination();
			logger.info("SettingsManager beendet");

			logger.info("Starte Finalisierung...");
			if(persistenceMgr != null && runningProperly) {
				persistenceMgr.prepareShutdown(dataIdentTree, new byte[1024 * 16]);
				persistenceMgr.deleteLockFile(); // Persistenzverzeichnis freigeben
			}
			logger.info("Finalisierung beendet (Indexe geschrieben, Lock-Datei geloescht)");
		}
		catch(Exception e) {
			logger.warning("Fehler beim Herunterfahren", e);
		}
		disconnectFromDav();

		persistenceMgr = null;
		dataIdentTree = null;
		inQueuesMgr = null;
		archivConfig = null;
		taskScheduler = null;
		simulationMgr = null;
	}

	/** Faehrt das ArS herunter und beendet die VM. */
	public void quit() {
		logger.info("Archivsystem wird beendet.");
		cleanUp();	// setzt terminated=true
		logger.info("Bye.");
		System.exit(0);
	}

	/** Faehrt das ArS herunter ohne die VM zu beenden (dies ist bei Fernsteuerung des ArS aus JUnit-Tests notwendig). */
	public void quitNoExit() {
		// Fuer den Fall, dass ein Test das ArS mehrfach beendet...
		if(!wasTerminated()) {
			logger.info("Archivsystem wird beendet - Java-VM laeuft weiter");
			cleanUp();	// setzt terminated=true
			logger.info("Bye.");
		}
	}

	/** Faehrt das ArS mit einer Fehlermeldung herunter und beendet die VM. */
	public void quitError(String msg) {
		logger.error(msg);
		logger.info("Archivsystem wird wegen eines Fehlers beendet.");
		cleanUp();
		logger.info("Bye (with error). " + msg);
		System.exit(0);
	}

	/** Faehrt das ArS mit einer Fehlermeldung herunter. */
	public void quitErrorNoExit(String msg) {
		logger.error(msg);
		logger.info("Archivsystem wird wegen eines Fehlers beendet.");
		cleanUp();
		logger.info("Bye (with error).");
	}

	/**
	 * Startpunkt des Archivsystems.
	 *
	 * @param args Kommandozeilenparameter
	 */
	public static void main(String[] args) {
		try {
			ArchiveManager am = new ArchiveManager(args);
			am.run();
		}
		catch(Exception ex) {
			ex.printStackTrace();
			System.exit(0);
		}
	}

	public ArchivConfig getArchivConfig() {
		return archivConfig;
	}

	public DataIdentTree getDataIdentTree() {
		return dataIdentTree;
	}

	public InQueuesMgr getInQueuesMgr() {
		return inQueuesMgr;
	}

	public PersistenceManager getPersMgr() {
		return persistenceMgr;
	}

	public TaskScheduler getTaskScheduler() {
		return taskScheduler;
	}


	/** @return Liefert den SimulationsManager. */
	public SimulationManager getSimulationMgr() {
		return simulationMgr;
	}

	/** @return Liefert die original Command Line Argumente als unmodifiableList von String Objekten. */
	public List<String> getOriginalCommandLineArguments() {
		List<String> result = new ArrayList<String>();
		for(String arg : argList.getInitialArgumentStrings()) {
			result.add(arg);
		}
		return Collections.unmodifiableList(result);
	}

	public int increaseArchiveQueryCountForApplication(SystemObject application) {
		synchronized(_applicationToArchiveQueryCount) {
			final Integer queryCount = _applicationToArchiveQueryCount.get(application);
			final int newCount;
			if(queryCount == null) {
				newCount = 1;
			}
			else {
				newCount = queryCount.intValue() + 1;
			}
			_applicationToArchiveQueryCount.put(application, Integer.valueOf(newCount));
			//System.out.println("increaseArchiveQueryCountForApplication: newCount = " + newCount);
			return newCount;
		}
	}

	public int decreaseArchiveQueryCountForApplication(SystemObject application) {
		synchronized(_applicationToArchiveQueryCount) {
			final Integer queryCount = _applicationToArchiveQueryCount.get(application);
			final int newCount;
			if(queryCount == null) {
				throw new IllegalStateException("Anzahl Archivanfragen fr Applikation " + application + " darf nicht kleiner als 0 werden");
			}
			else {
				newCount = queryCount.intValue() - 1;
			}
			if(newCount == 0) {
				_applicationToArchiveQueryCount.remove(application);
			}
			else {
				_applicationToArchiveQueryCount.put(application, Integer.valueOf(newCount));
			}
			//System.out.println("decreaseArchiveQueryCountForApplication: newCount = " + newCount);
			return newCount;
		}
	}
}
