/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.mgmt.datatree;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.dataIdentificationSettings.DataIdentification;
import de.bsvrz.sys.funclib.losb.IDMap;

import java.util.*;

/**
 * Hash-Tabelle, die pro Daten-Identifikation einen DataIdentNode haelt. Baumartige Implementierung als mehrstufige Hash-Map aus mehreren {@link
 * de.bsvrz.sys.funclib.losb.IDMap}.
 * <p/>
 * Derzeit werden keine DataIdentNodes geloescht.
 *
 * @author beck et al. projects GmbH
 * @author Thomas Schaefer
 * @version $Revision: 6419 $ / $Date: 2009-03-10 23:15:57 +0100 (Tue, 10 Mar 2009) $ / ($Author: rs $)
 */
public class DataIdentTree {

	/** Wurzel-Knoten der mehrstufigen Hash-Map. */
	private RootNode rootNode;

	/** Erzeugt einen neuen DataIdentTree. */
	public DataIdentTree() {
		rootNode = new RootNode();
	}

	/**
	 * Fuegt den uebergebenen DataIdentNode mit der angegebenen Daten-Identifikation ein.
	 *
	 * @param did Daten-Identifikation.
	 * @param din Abzulegender DataIdentNode.
	 */
	public synchronized void putDataIdentNode(DataIdentification did, DataIdentNode din) {
		DataDescription dd = did.getDataDescription();
		putDataIdentNode(did.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant(), din);
	}

	/**
	 * Fuegt den uebergebenen DataIdentNode mit der angegebenen Daten-Identifikation ein.
	 *
	 * @param objId Objekt-ID
	 * @param atgId Attributgruppen-ID
	 * @param aspId Aspekt-ID
	 * @param sv    Simulationsvariante
	 * @param din   Abzulegender DataIdentNode.
	 */
	public synchronized void putDataIdentNode(long objId, long atgId, long aspId, int sv, DataIdentNode din) {
		// 1. Objekt:
		ObjNode objNode = rootNode.getObjNode(objId);
		if(objNode == null) {
			objNode = new ObjNode();
			rootNode.putObjNode(objId, objNode);
		}
		// 2. Attributgruppe:
		AtgNode atgNode = objNode.getAtgNode(atgId);
		if(atgNode == null) {
			atgNode = new AtgNode();
			objNode.putAtgNode(atgId, atgNode);
		}
		// 3. Aspekt:
		AspNode aspNode = atgNode.getAspNode(aspId);
		if(aspNode == null) {
			aspNode = new AspNode();
			atgNode.putAspNode(aspId, aspNode);
		}
		// 4. DataIdentNode:
		aspNode.putDataIdentNode(sv, din);
	}

	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation.
	 *
	 * @param did Datenidentifikation
	 *
	 * @return DataIdentNode
	 *
	 * @throws DIdNodeNotFoundException falls nicht gefunden.
	 */
	public synchronized DataIdentNode getDataIdentNode(DataIdentification did) throws DIdNodeNotFoundException {
		DataDescription dd = did.getDataDescription();
		return getDataIdentNode(did.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}

	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation.
	 *
	 * @param rd Datensatz vom Datenverteiler
	 *
	 * @throws DIdNodeNotFoundException
	 * @return DataIdentNode zur angegebenen Daten-Identifikation.
	 */
	public synchronized DataIdentNode getDataIdentNode(ResultData rd) throws DIdNodeNotFoundException {
		DataDescription dd = rd.getDataDescription();
		return getDataIdentNode(rd.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}


	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation.
	 *
	 * @param ads Spezifikation des Archivdatensatzes (enthaelt objID, atgID, aspID, sv)
	 *
	 * @return DataIdentNode zur angegebenen Daten-Identifikation.
	 *
	 * @throws DIdNodeNotFoundException
	 */
	public synchronized DataIdentNode getDataIdentNode(ArchiveDataSpecification ads) throws DIdNodeNotFoundException {
		DataDescription dd = ads.getDataDescription();
		return getDataIdentNode(ads.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}


	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation.
	 *
	 * @param objID ID des Objektes
	 * @param atgID ID der Attributgruppe
	 * @param aspID ID des Aspektes
	 * @param sv    Simulationsvariante
	 *
	 * @return DataIdentNode
	 *
	 * @throws DIdNodeNotFoundException falls nicht gefunden.
	 */
	public synchronized DataIdentNode getDataIdentNode(long objID, long atgID, long aspID, int sv) throws DIdNodeNotFoundException {
		ObjNode objNode = rootNode.getObjNode(objID);
		if(objNode == null) throw new DIdNodeNotFoundException("ObjID " + objID + " nicht  vorhanden");
		AtgNode atgNode = objNode.getAtgNode(atgID);
		if(atgNode == null) throw new DIdNodeNotFoundException("AtgID " + atgID + " unter ObjID=" + objID + " nicht vorhanden");
		AspNode aspNode = atgNode.getAspNode(aspID);
		if(aspNode == null) throw new DIdNodeNotFoundException("AspID " + aspID + " unter ObjID " + objID + ", AtgID=" + atgID + " nicht vorhanden");
		DataIdentNode node = aspNode.getDataIdentNode(sv);
		if(node == null) {
			throw new DIdNodeNotFoundException("SimVar " + sv + " unter ObjID=" + objID + ", AtgID=" + atgID + ", AspID=" + aspID + " nicht vorhanden");
		}
		else {
			return node;
		}
	}

	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation und legt ihn vorher noetigenfalls an, falls er noch nicht existiert. Dieser DataIdentNode
	 * gehoert nicht zur aktuellen Archivparametrierung, sondern wurde z.B. aufgrund einer Anfrage um die Indexe zu speichern eingefuegt. Ggf. koennen solche
	 * DIN wieder geloescht werden, siehe cleanUp().
	 *
	 * @param did Datenidentifikation
	 *
	 * @return DataIdentNode
	 */
	public synchronized DataIdentNode existsCreate(DataIdentification did) {
		DataDescription dd = did.getDataDescription();
		return existsCreate(did.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}

	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation und legt ihn vorher noetigenfalls an, falls er noch nicht existiert. Dieser DataIdentNode
	 * gehoert nicht zur aktuellen Archivparametrierung, sondern wurde z.B. aufgrund einer Anfrage um die Indexe zu speichern eingefuegt. Ggf. koennen solche
	 * DIN wieder geloescht werden, siehe cleanUp().
	 *
	 * @param ads Spezifikation des Archivdatensatzes (enthaelt objID, atgID, aspID, sv)
	 *
	 * @return DataIdentNode
	 */
	public synchronized DataIdentNode existsCreate(ArchiveDataSpecification ads) {
		DataDescription dd = ads.getDataDescription();
		return existsCreate(ads.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}

	/**
	 * Liefert den DataIdentNode zur angegebenen Daten-Identifikation und legt ihn vorher noetigenfalls an, falls er noch nicht existiert. Dieser DataIdentNode
	 * gehoert nicht zur aktuellen Archivparametrierung, sondern wurde z.B. aufgrund einer Anfrage um die Indexe zu speichern eingefuegt. Ggf. koennen solche
	 * DIN wieder geloescht werden, siehe cleanUp().
	 *
	 * @param objId Objekt-ID
	 * @param atgId Attributgruppen-ID
	 * @param aspId Aspekt-ID
	 * @param sv    Simulationsvariante
	 *
	 * @return DataIdentNode
	 */
	public synchronized DataIdentNode existsCreate(long objId, long atgId, long aspId, int sv) {
		DataIdentNode din = null;
		try {
			din = getDataIdentNode(objId, atgId, aspId, sv);
		}
		catch(DIdNodeNotFoundException e) {
			// DIN existiert bisher noch nicht
		}
		if(din == null) {
			din = new DataIdentNode();
			putDataIdentNode(objId, atgId, aspId, sv, din);
		}
		return din;
	}

	/**
	 * Loescht den DataIdentNode zur angegebenen Daten-Identifikation. Nach dem Loeschen des DataIdentNodes leer gewordene ObjNodes, AtgNodes oder AspNodes
	 * werden ebenfalls aus dem Baum entfernt.
	 *
	 * @param did Datenidentifikation
	 *
	 * @throws DIdNodeNotFoundException
	 */
	public synchronized void deleteDataIdentNode(DataIdentification did) throws DIdNodeNotFoundException {
		DataDescription dd = did.getDataDescription();
		deleteDataIdentNode(did.getObject().getId(), dd.getAttributeGroup().getId(), dd.getAspect().getId(), dd.getSimulationVariant());
	}

	/**
	 * Loescht den DataIdentNode zur angegebenen Daten-Identifikation. Nach dem Loeschen des DataIdentNodes leer gewordene ObjNodes, AtgNodes oder AspNodes
	 * werden ebenfalls aus dem Baum entfernt.
	 *
	 * @param objID ID des Objektes
	 * @param atgID ID der Attributgruppe
	 * @param aspID ID des Aspektes
	 * @param sv    Simulationsvariante
	 *
	 * @throws DIdNodeNotFoundException
	 */
	public synchronized void deleteDataIdentNode(long objID, long atgID, long aspID, int sv) throws DIdNodeNotFoundException {
		// Check, ob DIN vorhanden ist; wirft DataIdentException, falls nicht vorhanden:
		getDataIdentNode(objID, atgID, aspID, sv);

		// Alle Zwischen-Knoten im Baum:
		ObjNode objNode = rootNode.getObjNode(objID);
		AtgNode atgNode = objNode.getAtgNode(atgID);
		AspNode aspNode = atgNode.getAspNode(aspID);

		// DataIdentNode zur SimVar loeschen:
		aspNode.removeDataIdentNode(sv);
		if(aspNode.isEmpty()) {
			// Leeren ASP-Knoten loeschen:
			atgNode.removeAspNode(aspID);
			if(atgNode.isEmpty()) {
				// Leeren ATG-Knoten loeschen:
				objNode.removeAtgNode(atgID);
				if(objNode.isEmpty()) {
					// Leeren OBJ-Knoten loeschen:
					rootNode.removeObjNode(objID);
				}
			}
		}
	}

	/**
	 * Liefert alle Objekte, die im rootNode gespeichert sind. ObjekteIds werden mit Hilfe des {@link DataModel} aufgelst.
	 *
	 * @param model das DataModel
	 *
	 * @return ein Array von Systemobjekten
	 *
	 * @throws ConfigurationException
	 */
	public SystemObject[] getObjects(DataModel model) throws ConfigurationException {
		HashMap<Long, ObjNode> objectMap = this.getRootNode().getObjMapCopy();
		Set keys = objectMap.keySet();
		int i = 0;
		SystemObject[] objIds = new SystemObject[keys.size()];
		for(Iterator iter = keys.iterator(); iter.hasNext();) {
			long objId = Long.parseLong("" + iter.next());
			objIds[i] = model.getObject(objId);
			i++;
		}
		return objIds;
	}

	/**
	 * Liefert alle Attributgruppen zu einem Objekt, die im Baum gespeichert sind. AtgIds werden mit Hilfe des {@link DataModel} aufgelst.
	 *
	 * @param objID
	 * @param model
	 *
	 * @return ein Array von Attributgruppen
	 *
	 * @throws ConfigurationException
	 */
	public AttributeGroup[] getAtgs(long objID, DataModel model) throws ConfigurationException {
		HashMap<Long, AtgNode> atgMap = this.getRootNode().getObjNode(objID).getAtgMapCopy();
		Set keys = atgMap.keySet();
		int i = 0;
		AttributeGroup[] atgs = new AttributeGroup[keys.size()];
		for(Iterator iter = keys.iterator(); iter.hasNext();) {
			long atgId = Long.parseLong("" + iter.next());
			atgs[i] = (AttributeGroup)model.getObject(atgId);
			i++;
		}
		return atgs;
	}

	/**
	 * Liefert alle mglichen Aspekte zu einem Objekt und einer zugehrigen Attributgruppe AspekteIds werden mit Hilfe des {@link DataModel} aufgelst.
	 *
	 * @param objId die Id des Objekts
	 * @param atgId die Id der Attributgruppe
	 * @param model das DataModel
	 *
	 * @return ein Array von Aspekten
	 *
	 * @throws ConfigurationException
	 */
	public Aspect[] getAsps(long objId, long atgId, DataModel model) throws ConfigurationException {
		HashMap<Long, AspNode> aspMap = this.getRootNode().getObjNode(objId).getAtgNode(atgId).getAspMapCopy();
		Set keys = aspMap.keySet();
		int i = 0;
		Aspect[] asps = new Aspect[keys.size()];
		for(Iterator iter = keys.iterator(); iter.hasNext();) {
			long aspId = Long.parseLong("" + iter.next());
			asps[i] = (Aspect)model.getObject(aspId);
			i++;
		}
		return asps;
	}

	/**
	 * Liefert alle mglichen Simulationsvarianten zu einem Objekt mit Attributgruppe und Aspekt. Werte werden direkt aus dem AspNode gelesen.
	 *
	 * @param objId die Id des Objekts
	 * @param atgId die Id der Attributgruppe
	 * @param aspId die Id des Aspekts
	 *
	 * @return ein Array von mglichen Simulationsvarianten
	 */
	public int[] getSimVars(long objId, long atgId, long aspId) {
		HashMap<Integer, DataIdentNode> identMap = this.getRootNode().getObjNode(objId).getAtgNode(atgId).getAspNode(aspId).getDinMapCopy();
		Set keys = identMap.keySet();
		int i = 0;
		int[] simVars = new int[keys.size()];
		for(Iterator iter = keys.iterator(); iter.hasNext();) {
			simVars[i] = Integer.parseInt("" + iter.next());
			i++;
		}
		return simVars;
	}

	/**
	 * Liefert einen DataIdentTreeIterator zum Iterieren des DataIdentTrees zurueck.
	 *
	 * @return DataIdentTreeIterator
	 */
	public DataIdentTreeIterator createIterator() {
		return new DataIdentTreeIterator(this);
	}

	/**
	 * Liefet die Anzahl der DataIdentNodes im Baum zurueck.
	 *
	 * @return Anzahl DataIdentNodes
	 */
	public int countDataIdentNodes() {
		int result = 0;
		DataIdentTreeIterator it = createIterator();
		while(it.getNext()) result++;
		return result;
	}

	/**
	 * Liefert einen Stringbuffer mit allen Datenidentifikation.
	 *
	 * @return Stringbuffer
	 */
	public StringBuilder print() {
		StringBuilder sb = new StringBuilder();
		DataIdentTreeIterator it = createIterator();
		while(it.getNext()) {
			sb.append("Datenidentifikation: Obj=" + it.getObjId() + " Atg=" + it.getAtgId() + " Asp=" + it.getAspId() + " SV=" + it.getSV() + "\n");
		}
		return sb;
	}

	/**
	 * Liefert den Wurzel-Knoten des DataIdentTree als Startpunkt fuer eine Iteration ueber den ganzen Baum.
	 *
	 * @return RootNode
	 */
	RootNode getRootNode() {
		return rootNode;
	}

	/** DataIdentNodes lschen, die nicht parametriert sind. */
	public void cleanUp() {
		DataIdentTreeIterator it = createIterator();

		while(it.getNext()) {
			DataIdentNode din = it.getDataIdentNode();
			boolean criticalAreaEntered = false;
			try {
				criticalAreaEntered = din.enterCriticalArea();
				if(!criticalAreaEntered) return; // Abbruch, da kritischer Bereich nicht zugaenglich wegen Task-Interrupt.

				if(!din.isArSParameterized()) {
					deleteDataIdentNode(it.getObjId(), it.getAtgId(), it.getAspId(), it.getSV());
				}
			}
			catch(DIdNodeNotFoundException dinfEx) {
			}
			finally {
				if(criticalAreaEntered) din.exitCriticalArea();
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Innere Klassen:
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	/** Ein Objekt diesen Typs stellt den Wurzel-Knoten der mehrstufigen Hashtabelle dar. Der RootNode haelt eine Map seiner Objekte nach Objekt-ID. */
	class RootNode {

		private IDMap<Long, ObjNode> objMap = new IDMap<Long, ObjNode>();

		void putObjNode(long objId, ObjNode objNode) {
			objMap.put(objId, objNode);
		}

		ObjNode getObjNode(long objId) {
			return objMap.get(objId);
		}

		void removeObjNode(long objId) {
			objMap.remove(objId);
		}

		boolean isEmpty() {
			return objMap.isEmpty();
		}

		HashMap<Long, ObjNode> getObjMapCopy() {
			return objMap.getCopy();
		}
	}

	/**
	 * Objekte diesen Typs stehen in der mehrstufigen Hashtabelle fr eine Objekt-Identifikation. Jeder ObjNode haelt eine Map seiner Attributgruppen nach
	 * Attributgruppen-ID.
	 */
	class ObjNode {

		private IDMap<Long, AtgNode> atgMap = new IDMap<Long, AtgNode>();

		void putAtgNode(long atgId, AtgNode atgNode) {
			atgMap.put(atgId, atgNode);
		}

		AtgNode getAtgNode(long atgId) {
			return atgMap.get(atgId);
		}

		void removeAtgNode(long atgId) {
			atgMap.remove(atgId);
		}

		boolean isEmpty() {
			return atgMap.isEmpty();
		}

		HashMap<Long, AtgNode> getAtgMapCopy() {
			return atgMap.getCopy();
		}
	}

	/**
	 * Objekte diesen Typs stehen in der mehrstufigen Hashtabelle fr eine Attributgruppen-Identifikation. Jeder AtgNode haelt eine Map seiner Aspekte nach
	 * Aspekt-ID.
	 */
	class AtgNode {

		private IDMap<Long, AspNode> aspMap = new IDMap<Long, AspNode>();

		void putAspNode(long aspId, AspNode aspNode) {
			aspMap.put(aspId, aspNode);
		}

		AspNode getAspNode(long aspId) {
			return aspMap.get(aspId);
		}

		void removeAspNode(long aspId) {
			aspMap.remove(aspId);
		}

		boolean isEmpty() {
			return aspMap.isEmpty();
		}

		HashMap<Long, AspNode> getAspMapCopy() {
			return aspMap.getCopy();
		}
	}

	/**
	 * Objekte diesen Typs stehen in der mehrstufigen Hashtabelle fr eine Aspekt-Identifikation. Jeder AspNode haelt eine Map seiner DataIdentNodes nach
	 * Simulations-Variante. Es wird eine platzsparende Implementierung verwendet.
	 */
	class AspNode {

		private Object didNodes = null;	// enhaelt entweder direkt den DataIdentNode (nur bei simVar==0) oder eine Map

		@SuppressWarnings({"unchecked"})
		void putDataIdentNode(int simVar, DataIdentNode din) {
			if(didNodes == null) {
				if(simVar == 0) {
					didNodes = din;
				}
				else {
					didNodes = new IDMap<Integer, DataIdentNode>();
					((IDMap)didNodes).put(simVar, din);
				}
			}
			else {
				if(didNodes instanceof DataIdentNode) {
					if(simVar != 0) {
						IDMap<Integer, DataIdentNode> map = new IDMap<Integer, DataIdentNode>();
						map.put(0, (DataIdentNode)didNodes);
						map.put(simVar, din);
						didNodes = map;
					}
				}
				else {
					((IDMap)didNodes).put(simVar, din);
				}
			}
		}

		@SuppressWarnings({"unchecked"})
		DataIdentNode getDataIdentNode(int simVar) {
			if(didNodes != null) {
				if(didNodes instanceof DataIdentNode) {
					if(simVar == 0) {
						return (DataIdentNode)didNodes;
					}
					else {
						return null;
					}
				}
				else {
					return ((IDMap<Integer, DataIdentNode>)didNodes).get(simVar);
				}
			}
			else {
				return null;
			}
		}

		@SuppressWarnings({"unchecked"})
		void removeDataIdentNode(int simVar) {
			if(didNodes != null) {
				if(didNodes instanceof DataIdentNode) {
					if(simVar == 0) didNodes = null;
				}
				else {
					((IDMap)didNodes).remove(simVar);
					// Falls nur noch simVar==0 drin ist, wieder direkt speichern
					if(((IDMap)didNodes).size() == 1 && ((IDMap)didNodes).containsKey(0)) didNodes = ((IDMap<Integer, DataIdentNode>)didNodes).get(0);
				}
			}
		}

		@SuppressWarnings({"unchecked"})
		boolean isEmpty() {
			if(didNodes != null) {
				if(didNodes instanceof DataIdentNode) {
					return false;
				}
				else {
					return ((IDMap<Integer, DataIdentNode>)didNodes).isEmpty();
				}
			}
			else {
				return true;
			}
		}

		@SuppressWarnings({"unchecked"})
		HashMap<Integer, DataIdentNode> getDinMapCopy() {
			if(didNodes != null) {
				if(didNodes instanceof DataIdentNode) {
					HashMap<Integer, DataIdentNode> hash = new HashMap<Integer, DataIdentNode>();
					hash.put(0, (DataIdentNode)didNodes);
					return hash;
				}
				else {
					return ((IDMap<Integer, DataIdentNode>)didNodes).getCopy();
				}
			}
			else {
				return new HashMap();
			}
		}
	}
    
}
