/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.ars.ars.
 * 
 * de.bsvrz.ars.ars is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.ars.ars is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.ars.ars; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.ars.ars.persistence.index;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.RandomAccessFile;

/**
 * Diese Klasse implementiert einen monoton steigenden Index. Jeder Eintrag des Index besteht aus Datenindex-Minimum, Datenindex-Maximum, Archivzeit-Minimum,
 * Archivzeit-Maximum, Datenzeit-Minimum, Datenzeit-Maximum und der Container-ID. In dieser Klasse ist gefordert, dass die Archivzeit monoton und die
 * ContainerID streng monoton ueber alle Eintraege hinweg steigend sind. Datenindex und Datenzeit duerfen Rueckspruenge aufweisen (benoetigt fuer nachgeforderte
 * Datensaetze). Daher kann diese Klasse nur nach Archivzeit suchen.
 * <p/>
 * Zudem muss in jedem Eintrag das Maximum groesser oder gleich dem Minimum sein.
 * <p/>
 * Wenn die Zahl der Eintrge die Pufferkapazitt im Speicher bersteigt, werden die Eintrge in die Indexdatei geschrieben. Je grer die Pufferkapazitt im
 * Speicher, umso schneller ist die Suche, weil keine Plattenzugriffe notwendig sind. Die Suche nach ContainerIDs ist als Binrsuche implementiert, die
 * automatisch im Speicher und in der Datei sucht.
 *
 * @author beck et al. projects GmbH
 * @author Phil Schrettenbrunner
 * @author Alexander Schmidt
 * @version $Revision: 8213 $ / $Date: 2010-09-28 17:00:20 +0200 (Tue, 28 Sep 2010) $ / ($Author: rs $)
 */
public class ATimeMonotoneIndex {

	protected static enum IndexStartmode {

		APPEND_REUSE_LAST,
		ERASE_OLD
	}

	;

	public static final IndexStartmode APPEND_REUSE_LAST = IndexStartmode.APPEND_REUSE_LAST;

	public static final IndexStartmode ERASE_OLD = IndexStartmode.ERASE_OLD;

	public static final int NO_VALUE = -1;

	/** Der Dateiname, unter dem alle Indices dieser Art gespeichert werden (der Speicherort wird spter bestimmt). */
	public static final String IDX_FILENAME = "_didx_atime.idx";

	public static enum Key {

		DI,
		AT,
		DT
	}

	;

	protected static final Key DI = Key.DI, AT = Key.AT, DT = Key.DT;

	private static final boolean MIN = true, MAX = false;

	public static final int ENTRY_LEN = IndexEntryMonotone.ENTRY_LEN;

	private byte[] byteArray;			// Feld fr N Eintrge

	private int rbFstEnt, rbLstEnt;		// Zeiger des Ringpuffers (aeltester/neuester Eintrag)

	private int rbFstUnsaved;			// ltester Eintrag (0..N-1), der nicht auf Platte ist

	private boolean ignoreLastFileEntry;	// gibt an ob der letzte Eintrag aufgrund von Startmode==APPEND_REUSE_LAST ueberschrieben und beim suchen in der Datei ignoriert werden soll.

	private boolean noActualEntry = true;

	/**
	 * Erzeugt einen neuen monoton steigenden Index.
	 * <p/>
	 * Bei startmode==APPEND_REUSE_LAST wird der Ringpuffer mit den aktuellsten Daten aus einer vorhandenen Indexdate gefllt. Zustzlich wird der letzte Eintrag
	 * in der Indexdatei zum currentEntry. Damit kann ein Container fortgesetzt und der Key2 weiter aktualisiert werden.
	 * <p/>
	 * Bei startmode==CREATE_NEW wird ein leerer Index erzeugt und die bisherige Indexdatei (sofern vorhanden) geloescht.
	 *
	 * @param size      Gre des Buffers (Anzahl der Eintrge)
	 * @param basePath  Verzeichnis der Indexdatei
	 * @param startmode {@link IndexStartmode#ERASE_OLD}: alte Indexdatei loeschen, {@link IndexStartmode#APPEND_REUSE_LAST}: anfuegen, letzten Eintrag
	 *                  weiterverwenden
	 *
	 * @throws IndexException
	 */
	public ATimeMonotoneIndex(int size, String basePath, IndexStartmode startmode) throws IndexException {
		this(size);
		if(startmode == IndexStartmode.ERASE_OLD) {
			File oldIndexFile = new File(basePath, IDX_FILENAME);
			if(oldIndexFile.exists() && !oldIndexFile.delete()) {
				throw new IndexException(
						"Fehler beim Neuanlegen eines Index: Alte Indexdatei '" + oldIndexFile.getAbsolutePath() + "' konnte nicht geloescht werden."
				);
			}
		}
		else if(startmode == IndexStartmode.APPEND_REUSE_LAST) reuseIndexFromFile(basePath);
	}

	/**
	 * Erzeugt einen neuen monoton steigenden Index. Neue Eintraege werden an eine bestehende Indexdatei angefuegt, wenn diese bereits existiert. Existiert keine
	 * Indexdatei, wird eine erzeugt. Initial werden keine Daten eingelesen; es findet kein Plattenzugriff statt.
	 *
	 * @param size Gre des Buffers (Anzahl der Eintrge). Wird mindestens auf 1 gesetzt.
	 */
	public ATimeMonotoneIndex(int size) {
		ignoreLastFileEntry = false;
		byteArray = new byte[(size <= 0 ? 1 : size) * ENTRY_LEN];
		rbFstEnt = 0;
		rbLstEnt = NO_VALUE;
		rbFstUnsaved = NO_VALUE;
	}


	/**
	 * Verndert die Gre des RingBuffers. Beim Verkleinern werden nicht gespeicherte Eintrge auf Platte geschrieben
	 *
	 * @param newSize
	 * @param basePath
	 *
	 * @throws IndexException
	 */
	public void resize(int newSize, String basePath) throws IndexException {
		assert newSize > 0 : "Neue Gre kleiner/gleich 0";
		newSize = newSize > 0 ? newSize : 1;

		if(capacity() == newSize)               // Kein Umkopieren ntig, da gleiche Gre
		{
			return;
		}
		if(entries() == 0) {                    //  Keine Daten -> nur neuen Buffer anlegen
			byteArray = new byte[newSize * ENTRY_LEN];
			return;
		}

		if(newSize > capacity()) {
			// Vergrern
			if(rbFstEnt <= rbLstEnt) {
				byte oldBuffer[] = byteArray;
				int start = rbFstEnt * ENTRY_LEN;
				int length = rbLstEnt * ENTRY_LEN - start + ENTRY_LEN;
				byteArray = new byte[newSize * ENTRY_LEN];
				System.arraycopy(oldBuffer, start, byteArray, start, length);
			}
			else {
				byte oldBuffer[] = byteArray;
				int num = entries();
				int start = rbFstEnt * ENTRY_LEN;
				int length = byteArray.length - start;
				int length2 = rbLstEnt * ENTRY_LEN + ENTRY_LEN;
				byteArray = new byte[newSize * ENTRY_LEN];
				System.arraycopy(oldBuffer, start, byteArray, 0, length);
				start = length;
				System.arraycopy(oldBuffer, 0, byteArray, start, length2);
				if(rbFstUnsaved != NO_VALUE) rbFstUnsaved = rbFstUnsaved >= rbFstEnt ? rbFstUnsaved - rbFstEnt : num - rbFstEnt + rbFstUnsaved;
				rbLstEnt = num - 1;
				rbFstEnt = 0;
			}
		}
		else {
			// Verkleinern
			if(rbFstUnsaved != NO_VALUE && ((capacity() + rbLstEnt - rbFstUnsaved) % capacity()) > newSize) flush(basePath);
			int firstToCopy = (newSize > entries()) ? rbFstEnt : (entries() + rbLstEnt + 1 - newSize) % entries();

			if(firstToCopy < rbLstEnt) {
				byte oldBuffer[] = byteArray;
				int start = firstToCopy * ENTRY_LEN;
				int length = rb2pos(entries() - 1) * ENTRY_LEN - start + ENTRY_LEN;
				byteArray = new byte[newSize * ENTRY_LEN];
				System.arraycopy(oldBuffer, start, byteArray, 0, length);
				if(rbFstUnsaved != NO_VALUE) rbFstUnsaved = (length / ENTRY_LEN - 1) - (rbLstEnt - firstToCopy);
				rbFstEnt = 0;
				rbLstEnt = length / ENTRY_LEN - 1;
			}
			else {
				byte oldBuffer[] = byteArray;
				int start = firstToCopy * ENTRY_LEN;
				int length = oldBuffer.length - start;
				byteArray = new byte[newSize * ENTRY_LEN];
				System.arraycopy(oldBuffer, start, byteArray, 0, length);
				start = length;
				length = rbLstEnt * ENTRY_LEN + ENTRY_LEN;
				System.arraycopy(oldBuffer, 0, byteArray, start, length);
				rbLstEnt = (start + length) / ENTRY_LEN - 1;
				rbFstEnt = 0;
				if(rbFstUnsaved != NO_VALUE) rbFstUnsaved = rbLstEnt - rb2pos(rbFstUnsaved) - 1;
			}
		}
	}


	/**
	 * Fllt den Buffer so weit es geht mit den Daten aus der Indexdatei. Dabei wird nur der neuste Teil geladen, wenn die Datei mehr Daten hlt als in den Buffer
	 * passen. Der letzte Eintrag wird weiterverwendet und deswegen aus der Indexdatei gelscht.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 *
	 * @throws IndexException        Wenn es Probleme mit dem Einlesen der Datei gibt.
	 * @throws CorruptIndexException Wenn die Eingelesene Indexdatei beschdigt ist.
	 */
	private void reuseIndexFromFile(String basePath) throws IndexException, CorruptIndexException {
		File idxFile = checkIndexFile(basePath);
		if(!idxFile.exists()) {
			throw new CorruptIndexException("Weiterfhren der IndexDatei '" + idxFile.getAbsolutePath() + "' nicht mglich. Datei existiert nicht.");
		}

		readFile(basePath);
		rbFstUnsaved = rbLstEnt;					// Letzten Eintrag "dirty" machen...
		ignoreLastFileEntry = true;
		noActualEntry = false;
	}

	/**
	 * Fllt den Buffer so weit es geht mit den Daten aus der Indexdatei.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 *
	 * @throws IndexException Wenn es Probleme mit dem Einlesen der Datei gibt.
	 */
	private void readFile(String basePath) throws IndexException {
		File idxFile = checkIndexFile(basePath);
		if(idxFile.exists()) {
			RandomAccessFile raf = null;
			try {
				raf = new RandomAccessFile(idxFile, "r");
				raf.seek(raf.length() > byteArray.length ? raf.length() - byteArray.length : 0);
				int numRead = raf.read(byteArray);
				if(numRead <= 0) {
					throw new CorruptIndexException("Leere Indexdatei bei Weiterverwendung");
				}
				else {
					if(numRead % ENTRY_LEN != 0) {
						throw new CorruptIndexException("Lnge der Index-Datei (" + numRead + " Bytes) kein Vielfaches von " + ENTRY_LEN);
					}
					rbFstEnt = 0;
					rbLstEnt = numRead / ENTRY_LEN - 1;
				}
			}
			catch(IOException e) {
				throw new IndexException("Fehler beim Einlesen der Indexdatei '" + idxFile.getAbsolutePath() + "': " + e.getMessage(), e);
			}
			finally {
				closeRandomAccessFile(raf);
			}
		}
	}


	/**
	 * @param pos Indexposition (0 = neuester Eintrag, gezaehlt in Eintraegen)
	 *
	 * @return Container-ID an der angegebenen Indexposition
	 */
	private long getCID(int pos) {
		return entries() <= 0 ? NO_VALUE : IndexEntryMonotone.getCID(byteArray, (rbFstEnt + pos) % capacity());
	}

	/**
	 * Liefert den Key1 (Min) an der angegebenen Indexposition.
	 *
	 * @param entrKey
	 * @param min
	 * @param pos     Indexposition (0 = neuester Eintrag, gezaehlt in Eintraegen)
	 *
	 * @return Key1 der Indexposition
	 */
	private long getEntry(Key entrKey, boolean min, int pos) {
		return entries() <= 0 ? NO_VALUE : IndexEntryMonotone.get(entrKey, min, byteArray, (rbFstEnt + pos) % capacity());
	}

	/**
	 * Setzt einen Eintrag an der angegebenen Indexposition.
	 *
	 * @param pos   Indexposition (0 = neuester Eintrag, gezaehlt in Eintraegen)
	 * @param dIMin zu setzendes Datenindex-Minimum
	 * @param dIMax zu setzendes Datenindex-Maximum
	 * @param aTMin zu setzendes Archivzeit-Minimum
	 * @param aTMax zu setzendes Archivzeit-Maximum
	 * @param dTMin zu setzendes Datenzeit-Minimum
	 * @param dTMax zu setzendes Datenzeit-Maximum
	 * @param cID   zu setzende Container-ID
	 *
	 * @throws IndexException
	 */
	private void setEntry(int pos, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) throws IndexException {
		final int capacity = capacity();
		if(pos < 0 || pos >= capacity) throw new IndexException("setEntry() fr Position " + pos + " bei Capacity=" + capacity + " aufgerufen");
		IndexEntryMonotone.setEntry(byteArray, (rbFstEnt + pos) % capacity, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
	}

	/**
	 * Aktualisiert die Maximumwerte des aktuellsten Eintrages. Die Minimumwerte und die Container-ID bleiben unverndert.
	 *
	 * @param dIMax zu setzendes Datenindex-Maximum
	 * @param aTMax zu setzendes Archivzeit-Maximum
	 * @param dTMax zu setzendes Datenzeit-Maximum
	 *
	 * @throws IndexException
	 */
	public void updateMax(long dIMax, long aTMax, long dTMax) throws IndexException {
		long cDIMax = getCurrentDIMax(), cATMax = getCurrentATMax(), cDTMax = getCurrentDTMax();
		if(entries() <= 0) {
			throw new IndexException("updateMax() auf leerem Index aufgerufen");
		}
		else if(cDIMax >= dIMax) {
			throw new CorruptIndexException("neuer Wert nicht streng monoton steigend: currentDIdxMax=" + cDIMax + " newDIdxMax=" + dIMax);
		}
		else if(cATMax > aTMax) {
			throw new CorruptIndexException("neuer Wert nicht monoton steigend: currentATimeMax=" + cATMax + " newATimeMax=" + aTMax);
		}
		else if(cDTMax > dTMax) throw new CorruptIndexException("neuer Wert nicht monoton steigend: currentDTimeMax=" + cDTMax + " newDTimeMax=" + dTMax);

		setEntry(entries() - 1, getCurrentDIMin(), dIMax, getCurrentATMin(), aTMax, getCurrentDTMin(), dTMax, getCurrentCID());
	}

	/** @return ContainerID des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentCID() {
		return noActualEntry ? NO_VALUE : getCID(entries() - 1);
	}

	/** @return Minimum des Datenindex des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentDIMin() {
		return noActualEntry ? NO_VALUE : getEntry(DI, MIN, entries() - 1);
	}

	/** @return Maximum des Datenindex des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentDIMax() {
		return noActualEntry ? NO_VALUE : getEntry(DI, MAX, entries() - 1);
	}

	/** @return Minimum der Archivzeit des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentATMin() {
		return noActualEntry ? NO_VALUE : getEntry(AT, MIN, entries() - 1);
	}

	/** @return Maximum der Archivzeit des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentATMax() {
		return noActualEntry ? NO_VALUE : getEntry(AT, MAX, entries() - 1);
	}

	/** @return Minimum der Datenzeit des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentDTMin() {
		return noActualEntry ? NO_VALUE : getEntry(DT, MIN, entries() - 1);
	}

	/** @return Maximum der Datenzeit des aktuellen Eintrags oder -1 falls kein Eintrag vorhanden ist */
	public long getCurrentDTMax() {
		return noActualEntry ? NO_VALUE : getEntry(DT, MAX, entries() - 1);
	}

	/**
	 * Fgt einen neuen Eintrag in den Buffer ein. Bei Bedarf wird der Puffer auf Platte geschrieben.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param dIMin    Min-Wert des Datenindex
	 * @param dIMax    Max-Wert des Datenindex
	 * @param aTMin    Min-Wert der Archivzeit
	 * @param aTMax    Max-Wert der Archivzeit
	 * @param dTMin    Min-Wert der Datenzeit
	 * @param dTMax    Max-Wert der Datenzeit
	 * @param cID      Container-ID des Neueintrags
	 *
	 * @throws IndexException
	 */
	public void add(String basePath, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) throws IndexException {
		// Neue Werte pruefen:
		checkForAdd(basePath, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
		if(entries() == capacity()) {
			if(stepRB(rbLstEnt) == rbFstUnsaved) flush(basePath);
			rbFstEnt = stepRB(rbFstEnt);
		}
		rbLstEnt = stepRB(rbLstEnt);
		setEntry(entries() - 1, dIMin, dIMax, aTMin, aTMax, dTMin, dTMax, cID);
		if(rbFstUnsaved == NO_VALUE) rbFstUnsaved = rbLstEnt;
		noActualEntry = false;
	}

	/**
	 * Prft, ob dem Index ein neuer Eintrag hinzugefgt werden knnte.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param dIMin    Min-Wert des Datenindex
	 * @param dIMax    Max-Wert des Datenindex
	 * @param aTMin    Min-Wert der Archivzeit
	 * @param aTMax    Max-Wert der Archivzeit
	 * @param dTMin    Min-Wert der Datenzeit
	 * @param dTMax    Max-Wert der Datenzeit
	 * @param cID      Container-ID des Neueintrags
	 *
	 * @throws IndexException
	 */
	public void checkForAdd(String basePath, long dIMin, long dIMax, long aTMin, long aTMax, long dTMin, long dTMax, long cID) throws IndexException {
		// Neue Werte pruefen:
		try {
			if(dIMin < 0 || dIMax < 0 || aTMin < 0 || aTMax < 0 || dTMin < 0 || dTMax < 0 || cID <= 0 || dIMin > dIMax || aTMin > aTMax || dTMin > dTMax) {
				throw new Exception("Neueintrag im monotonen Index ungltig:");
			}
			else if(entries() > 0) {
				if(getCurrentATMax() > aTMin) {
					throw new Exception("Neueintrag nicht monoton: currentATimeMax=" + getCurrentATMax());
				}
				else if(getCurrentCID() >= cID) throw new Exception("Neueintrag nicht streng monoton: currentCId=" + getCurrentCID());
			}
		}
		catch(Exception e) {
			throw new CorruptIndexException(
					e.getMessage() + " (" + "dIMin=" + dIMin + ", dIMax=" + dIMax + ", aTMin=" + aTMin + ", aTMax=" + aTMax + ", dTMin=" + dTMin + ", dTMax="
					+ dTMax + ", contID=" + cID + ")",
					e
			);
		}
	}

	/**
	 * Fhrt einen Schritt im Ringpuffer aus (und bricht dabei an der Feldgrenze um).
	 *
	 * @param rbPos Absolute Position im Feld (gezhlt in Eintrgen).
	 *
	 * @return Nchste absolute Position im Feld (gezhlt in Eintrgen).
	 */
	private int stepRB(int rbPos) {
		return (rbPos + 1) % capacity();
	}

	/**
	 * Rechnet eine absolute Position im Feld (gezhlt in Eintrgen) in eine Ringpuffer-Position um (0 = neuester Eintrag).
	 *
	 * @param rbPos absolute Position im Feld (gezhlt in Eintrgen).
	 *
	 * @return Ringpuffer-Position (0 = neuester Eintrag)
	 */
	private int rb2pos(int rbPos) {
		if(rbPos < 0) {
			return -1;
		}
		else {
			return rbPos >= rbFstEnt ? rbPos - rbFstEnt : capacity() - rbFstEnt + rbPos;
		}
	}

	/**
	 * Liefert die IDs der Container, bei denen der gesuchte Key zwischen Anfang und Ende liegt.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param temp     Speicherbereich, der als Puffer dient, wenn auf die Datei zurckgegriffen werden muss.
	 * @param cmpKey
	 * @param minKey
	 * @param maxKey
	 *
	 * @return Liste von ContainerIDs mit Min/Max-Werten von DI, AT, DT
	 *
	 * @throws IndexException
	 */
	protected IndexResult getContainerID(String basePath, byte[] temp, Key cmpKey, long minKey, long maxKey) throws IndexException {
		assert cmpKey != DT : "Anfrage nach Datenzeit nicht moeglich im monotonen Index";
		if(minKey > maxKey) throw new IndexException("Fehlerhafte Suchparameter: min (" + minKey + ") > max (" + maxKey + ")");

		// Wenn der Puffer leer ist und wir sowieso aus der Datei lesen muessen, koennen wir auch gleich
		// den Puffer fuellen. Kann bei weiteren Suchvorgaengen einen Vorteil geben.
		if(entries() == 0) readFile(basePath);

		int lowPos = binaryMemSearch(cmpKey, minKey);
		if(lowPos > 0) {
			// komplett im Speicherindex, nach rechts suchen
			int anz = countMatchingEntries(lowPos, entries(), cmpKey, maxKey);
			IndexResult res = new IndexResult();
			for(int i = lowPos; i < lowPos + anz; i++) {
				res.add(
						getCID(i),
						getEntry(DI, MIN, i),
						getEntry(DI, MAX, i),
						getEntry(AT, MIN, i),
						getEntry(AT, MAX, i),
						getEntry(DT, MIN, i),
						getEntry(DT, MAX, i)
				);
			}
			return res;
		}
		else {
			return binaryFileSearch(basePath, temp, cmpKey, minKey, maxKey);
		}
	}

	/**
	 * Liefert die IDs der Container, deren Archivzeitspanne sich mit der angegebenen Spanne ueberschneidet.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param temp     Speicherbereich, der als Puffer dient, wenn auf die Datei zurckgegriffen werden muss.
	 * @param minATime Beginn der Zeitspanne
	 * @param maxATime Ende der Zeitspanne
	 *
	 * @return Liste von ContainerIDs mit Min/Max-Werten von DI, AT, DT
	 *
	 * @throws IndexException
	 */
	public IndexResult getContainerIDByATime(String basePath, byte[] temp, long minATime, long maxATime) throws IndexException {
		return getContainerID(basePath, temp, AT, minATime, maxATime);
	}

	/**
	 * Liefert den kleinsten Archivzeit-Minwert, der jemals in diesem Index gespeichert wird.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 *
	 * @return Kleinster Archivzeitwert
	 *
	 * @throws IndexException Bei Zugriffsproblemen mit der IndexDatei
	 */
	public long getAbsoluteATMin(String basePath) throws IndexException {
		IndexResult ir = getMinEntry(basePath);
		if(ir != null && ir.size() > 0) {
			return ir.getATMin(0);
		}
		else {
			return IndexResult.NO_VALUE;
		}
	}

	/**
	 * Liefert den kleinsten Eintrag bei Sortierung nach Archivzeit.
	 *
	 * @param basePath Verzeichnis, in dem die Index-Datei liegt
	 *
	 * @return Kleinster Eintrag oder -1 falls keiner vorhanden
	 *
	 * @throws IndexException Bei Zugriffsproblemen mit der IndexDatei
	 */
	protected IndexResult getMinEntry(String basePath) throws IndexException {
		IndexResult ir = new IndexResult();
		File f = new File(basePath, IDX_FILENAME);

		if(f.length() == 0) {
			if(entries() == 0) {
				// Fall 1.1: Datei leer, Ram leer           => Leere Antwort
				return ir;
			}
			else {
				// Fall 1.2: Datei leer, Ram hlt Daten     => Antwort ist ltestes Element im Ram
				ir.add(
						getCID(0),
						getEntry(DI, MIN, 0),
						getEntry(DI, MAX, 0),
						getEntry(AT, MIN, 0),
						getEntry(AT, MAX, 0),
						getEntry(DT, MIN, 0),
						getEntry(DT, MAX, 0)
				);
				return ir;
			}
		}
		else {
			if(entries() == 0) {
				// Fall 2.1: Datei hlt Daten, Ram leer     => Daten laden. Antwort ist ltestes Element im Ram, wenn
				//                                             alle Daten ins Ram geladen werden konnten. Ansonsten wie 2.2 behandeln
				readFile(basePath);
				if(entries() * ENTRY_LEN == f.length()) {
					ir.add(
							getCID(0),
							getEntry(DI, MIN, 0),
							getEntry(DI, MAX, 0),
							getEntry(AT, MIN, 0),
							getEntry(AT, MAX, 0),
							getEntry(DT, MIN, 0),
							getEntry(DT, MAX, 0)
					);
					return ir;
				}
			}
			//    Fall 2.2: Daten und Ram enthlt Daten     => ltestes Element aus der Datei nehmen
			RandomAccessFile raf = null;
			try {
				raf = new RandomAccessFile(f, "rw");
				byte tmp[] = new byte[ENTRY_LEN];
				raf.seek(0);
				int bytesRead = raf.read(tmp, 0, tmp.length);
				if(bytesRead == ENTRY_LEN) {
					ir.add(
							IndexEntryMonotone.getCID(tmp, 0), IndexEntryMonotone.getDIdxMin(tmp, 0), IndexEntryMonotone.getDIdxMax(tmp, 0), IndexEntryMonotone
							.getATimeMin(tmp, 0), IndexEntryMonotone.getATimeMax(tmp, 0), IndexEntryMonotone.getDTimeMin(tmp, 0), IndexEntryMonotone
							.getDTimeMax(tmp, 0)
					);
				}
			}
			catch(IOException e) {
				throw new IndexException("Fehler beim Einlesen des kleinsten Eintrages: " + e.getMessage(), e);
			}
			finally {
				closeRandomAccessFile(raf);
			}
			return ir;
		}
	}

	/**
	 * Sucht den Puffer beginnend bei startAt in Richtung neuesten Eintrag nach Eintrgen ab, die zum Vergleichswert passen.
	 *
	 * @param startAt      Startposition
	 * @param numOfEntries Maximal zu durchsuchende Zahl von Eintrgen
	 * @param cmpKey       Vergleichswert des Eintrags
	 * @param key          Vergleichsschluessel
	 *
	 * @return Anzahl der passenden Eintrge
	 *
	 * @throws IndexException
	 */
	private int countMatchingEntries(int startAt, int numOfEntries, Key cmpKey, long key) throws IndexException {
		int pos = startAt - 1;
		while(++pos < numOfEntries && compare(cmpKey, pos, key) >= 0) ;
		return pos - startAt;
	}

	/** @return Liefert die Zahl der Eintrge, die im Speicher gehalten werden knnen. */
	public int capacity() {
		return byteArray.length / ENTRY_LEN;
	}

	/** @return Anzahl der Eintrge, die aktuell im Speicher gehalten werden. */
	public int entries() {
		return rbLstEnt == -1 ? 0 : (rbFstEnt > rbLstEnt ? capacity() - rbFstEnt + rbLstEnt + 1 : rbLstEnt - rbFstEnt + 1);
	}

	/**
	 * Sucht im Speicher nach dem ersten Eintrag, der auf den bergebenen Key passt. Dies ist der lteste Eintrag, bei dem Key gleich oder zwischen Key1 und Key2
	 * ist. Liegt Key zwischen zwei Eintrgen, wird der ltere genommen.
	 *
	 * @param cmpKey Vergleichswert des Eintrags
	 * @param key    Startschlssel
	 *
	 * @return Position des gefundenen Eintrags
	 *
	 * @throws IndexException
	 */
	private int binaryMemSearch(Key cmpKey, long key) throws IndexException {
		if(entries() == 0 || compare(cmpKey, 0, key) < 0) return NO_VALUE;	// leer oder nicht im Speicher

		int low = 0, high = entries() - 1, mid, cmp;
		while(low <= high) {
			mid = (low + high) / 2;
			cmp = compare(cmpKey, mid, key);
			if(cmp < 0) {
				high = mid - 1;
			}
			else if(cmp > 0) {
				low = mid + 1;
			}
			else {
				while(--mid >= 0 && compare(cmpKey, mid, key) == 0) ;	// Direkter Treffer, linkesten Eintrag suchen
				return mid + 1;
			}
		}
		// Kein direkter Treffer, sondern zwischen Eintrgen oder auerhalb des Index
		if(low >= entries()) {
			return entries() - 1;	// letzten Eintrag verwenden
		}
		else if(high < 0) {
			return -1;			// key kleiner ist als kleinster Eintrag
		}
		else {
			return high;		// zwischen zwei Eintrgen, linken nehmen
		}
	}

	/**
	 * Vergleicht den Eintrag an der gegebenen Position im Index mit dem gegebenen Schlssel.
	 *
	 * @param cmpKey Vergleichswert des Eintrags
	 * @param index  Position im Index (0 = neueste)
	 * @param key    Vergleichsschlssel
	 *
	 * @throws IndexException
	 * @return	-1 falls key < Key1(index); 1 falls key > Key2(index); 0 sonst
	 */
	private int compare(Key cmpKey, int index, long key) throws IndexException {
		return key < getEntry(cmpKey, MIN, index) ? -1 : (key > getEntry(cmpKey, MAX, index) ? 1 : 0);
	}

	/**
	 * Fhrt eine binrSuche in der IndexDatei durch. Dabei wird die IndexDatei in kleine Teile zerlegt, die dann in den bergebenen temporren Buffer eingelesen
	 * und mit dem gesuchten Wert verglichen werden.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 * @param tmpBuf   Ein temporres Array, das als Buffer verwendet wird.
	 * @param minKey   Schlssel, der gesucht wird
	 *
	 * @return Ein Array mit den gesuchten Werten
	 *
	 * @throws IndexException Wenn es Probleme mit der IndexDatei gibt.
	 */
	private IndexResult binaryFileSearch(String basePath, byte[] tmpBuf, Key cmpKey, long minKey, long maxKey) throws IndexException {
//    	assert tmpBuf.length >= 2*ENTRY_LEN : "Puffer kleiner als zweifache Lnge eines Eintrags";

		if(tmpBuf.length < 2 * ENTRY_LEN) {
			tmpBuf = new byte[2 * ENTRY_LEN];	// wenn kleinster/groesster wert unterschiedlich sind, ist das vergleichen einfacher
		}

		IndexResult res = new IndexResult();
		RandomAccessFile raf = null;
		File idxFile = checkIndexFile(basePath);
		if(idxFile.exists() && idxFile.length() > 0) {
			try {
				int tmpBufSize = (tmpBuf.length / ENTRY_LEN) * ENTRY_LEN;
				int totalChunks = (((int)idxFile.length()) / tmpBufSize) + (idxFile.length() % tmpBufSize == 0 ? 0 : 1);
				int chunk = 0, low = 0, high = totalChunks - 1, entriesRead = 0, compare = 0;
				raf = new RandomAccessFile(idxFile, "r");
				while(low <= high) {
					chunk = (low + high) / 2;
					entriesRead = readSearchChunk(raf, chunk, tmpBuf, tmpBufSize);
					if(entriesRead == 0) {
						// Wenn der Index nur einen Entrage hatte und dieser weiterverwendet wurde, wurde dieser bei readSearchChunk()
						// ignoriert und entriesRead==0. In diesem Fall muss im Speicher weitergesucht werden, was durch compare==1
						// erzwungen wird.
						compare = 1;
						break;
					}
					compare = minKey < IndexEntryMonotone.get(cmpKey, MIN, tmpBuf, 0)
					          ? -1
					          : (minKey > IndexEntryMonotone.get(cmpKey, MAX, tmpBuf, entriesRead - 1) ? 1 : 0);
					if(compare < 0) {
						high = chunk - 1;
					}
					else if(compare > 0) {
						low = chunk + 1;
					}
					else if(minKey == IndexEntryMonotone.get(cmpKey, MIN, tmpBuf, 0)) {
						// Treffer an der linkesten Kante, deshalb auch links anschlieende Chunks durchsuchen
						while(--chunk >= 0) {
							entriesRead = readSearchChunk(raf, chunk, tmpBuf, tmpBufSize);
							int subCompare = minKey < IndexEntryMonotone.get(cmpKey, MIN, tmpBuf, 0)
							                 ? -1
							                 : (minKey > IndexEntryMonotone.get(cmpKey, MAX, tmpBuf, entriesRead - 1) ? 1 : 0);
							if(subCompare != 0) break;
						}
						entriesRead = readSearchChunk(raf, ++chunk, tmpBuf, tmpBufSize);  // Chunk neu einlesen
						break;
					}
					else {
						break;
					}
				}

				if(!(chunk + 1 == totalChunks && compare == 1)) {
					// nicht groesser als groesster in der Datei, also Start im Chunk suchen

					// Indexzustand speichern, temporaeres Feld fr Binaersuche setzen
					int oldFstEnt = rbFstEnt, oldLstEnt = rbLstEnt;
					byte[] oldArray = byteArray;
					rbFstEnt = 0;
					rbLstEnt = entriesRead - 1;
					byteArray = tmpBuf;

					// Beginnend beim Start-Chunk Indexdatei nach rechts absuchen
					int startEntry = (chunk == 0 && compare == -1) ? 0 : binaryMemSearch(cmpKey, minKey);  // Binaersuche sparen falls kleiner als kleinster
					while(chunk++ < totalChunks) {
						int matchingEntr = countMatchingEntries(startEntry, Math.min(entries(), entriesRead), cmpKey, maxKey);
						if(matchingEntr == 0) break;
						for(int i = startEntry; i < startEntry + matchingEntr; i++) {
							res.add(
									getCID(i),
									getEntry(DI, MIN, i),
									getEntry(DI, MAX, i),
									getEntry(AT, MIN, i),
									getEntry(AT, MAX, i),
									getEntry(DT, MIN, i),
									getEntry(DT, MAX, i)
							);
						}
						startEntry = 0;
						entriesRead = readSearchChunk(raf, chunk, tmpBuf, tmpBufSize);
					}
					// Indexzustand wiederherstellen
					rbFstEnt = oldFstEnt;
					rbLstEnt = oldLstEnt;
					byteArray = oldArray;
				}

				// Indexteil im Speicher ab erstem ungesicherten (sofern vorhanden) durchsuchen
				if(rbFstUnsaved != NO_VALUE) {
					int fstUnsaved = rb2pos(rbFstUnsaved);
					int matchingEntr = countMatchingEntries(fstUnsaved, entries(), cmpKey, maxKey);
					for(int i = fstUnsaved; i < fstUnsaved + matchingEntr; i++) {
						res.add(
								getCID(i),
								getEntry(DI, MIN, i),
								getEntry(DI, MAX, i),
								getEntry(AT, MIN, i),
								getEntry(AT, MAX, i),
								getEntry(DT, MIN, i),
								getEntry(DT, MAX, i)
						);
					}
				}
			}
			catch(IOException e) {
				throw new IndexException("Fehler beim Suchen in der Indexdatei '" + idxFile.getAbsolutePath() + "': " + e.getMessage(), e);
			}
			finally {
				closeRandomAccessFile(raf);
			}
		}
		else { // keine Indexdatei vorhanden, Ergebnisse nur aus dem Speicher
			int anz = countMatchingEntries(0, entries(), cmpKey, maxKey);
			for(int i = 0; i < anz; i++) {
				res.add(
						getCID(i),
						getEntry(DI, MIN, i),
						getEntry(DI, MAX, i),
						getEntry(AT, MIN, i),
						getEntry(AT, MAX, i),
						getEntry(DT, MIN, i),
						getEntry(DT, MAX, i)
				);
			}
		}
		return res;
	}

	/**
	 * Liest einen Chunk zum Suchen in den Speicher. Falls der letzte Eintrag wegen Startmode=={@link IndexStartmode#APPEND_REUSE_LAST} weiterverwendet wird, wird
	 * der letzte Dateieintrag ignoriert.
	 *
	 * @param raf     Indexdatei
	 * @param chunkNo Nummer des Chunks, beginnend bei 0
	 * @param buf     Speicherbereich
	 * @param bufSize Gre des Speicherbereichs in Byte (ist stets durch entryLen() teilbar)
	 *
	 * @return Anzahl der eingelesenen Eintrge.
	 *
	 * @throws IOException
	 */
	private int readSearchChunk(RandomAccessFile raf, int chunkNo, byte[] buf, int bufSize) throws IOException {
		raf.seek(chunkNo * bufSize);
		int bytesRead = raf.read(buf, 0, bufSize);

		// am Dateiende und letzten weiterverwendet?
		if((chunkNo * bufSize + bytesRead == raf.length()) && ignoreLastFileEntry)
		// ...dann letzten Eintrag fuer die Suche ignorieren
		{
			return Math.max(bytesRead - ENTRY_LEN, 0) / ENTRY_LEN;
		}
		else {
			return bytesRead / ENTRY_LEN;
		}
	}

	/**
	 * Prft grundlegende Eigenschaften einer Indexdatei.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 *
	 * @return Indexdatei
	 *
	 * @throws IndexException
	 */
	private File checkIndexFile(String basePath) throws IndexException {
		File indexFile = new File(basePath, IDX_FILENAME);
		if(indexFile.exists()) checkFileLength(indexFile, indexFile.length());
		return indexFile;
	}

	private void checkFileLength(final File indexFile, final long fileLength) throws CorruptIndexException {
		if(fileLength % ENTRY_LEN != 0) {
			throw new CorruptIndexException("Die IndexDatei '" + indexFile.getAbsolutePath() + "' hat eine fehlerhafte Lnge.");
		}
	}

	/**
	 * Speichert den aktuellen Buffer in die Datei. Dabei werden nur die noch nicht gespeicherten Daten an die Daten in der Datei angehngt. Falls der Index mit
	 * {@link IndexStartmode#APPEND_REUSE_LAST} gestartet wurde, wird beim ersten Flushen der letzte Dateieintrag ueberschrieben.
	 *
	 * @param basePath Verzeichnis der Indexdatei
	 *
	 * @throws IndexException Wenn es Probleme mit der IndexDatei gibt
	 */
	private void flush(String basePath) throws IndexException {
		if(rbFstUnsaved != NO_VALUE) {
			File indexFile = new File(basePath, IDX_FILENAME);
			long fileLength = 0;
	
			if(indexFile.exists()) {
				fileLength = indexFile.length();
				checkFileLength(indexFile, fileLength);
			}

			//Datei wird erzeugt, falls sie nicht existiert
			try {
				int startByte = rbFstUnsaved % capacity() * ENTRY_LEN;
				int endByte = rbLstEnt % capacity() * ENTRY_LEN + ENTRY_LEN;

				byte[] bytesToWrite;

				// Wenn Puffer voll, dann ist startByte == endByte. In der ursprnglichen Version von bea wurde in diesem fall kein einziges Byte anstatt alle
				// Bytes des Puffers geschrieben.
				if(startByte >= endByte) {
					bytesToWrite = new byte[byteArray.length - startByte + endByte];

					System.arraycopy(byteArray, startByte, bytesToWrite, 0, byteArray.length - startByte);   // Hinterer Teil
					System.arraycopy(byteArray, 0, bytesToWrite, byteArray.length - startByte, endByte);     // Vorderer Teil

					startByte = 0;
					endByte = bytesToWrite.length;
				}
				else {
					bytesToWrite = byteArray;
				}

				if(ignoreLastFileEntry) {
					RandomAccessFile raf = new RandomAccessFile(indexFile, "rw");
					try {
						// ans Ende springen, bzw. eins davor, falls letzter Eintrag weiterverwendet wurde:
						raf.seek(Math.max(fileLength - ENTRY_LEN, 0));
						raf.write(bytesToWrite, startByte, endByte - startByte);
					}
					finally {
						closeRandomAccessFile(raf);
					}
					ignoreLastFileEntry = false;
				}
				else {
					final FileOutputStream os = new FileOutputStream(indexFile, true);
					try {
						os.write(bytesToWrite, startByte, endByte - startByte);
					}
					finally {
						os.close();
					}
				}
				rbFstUnsaved = NO_VALUE;
			}
			catch(IOException e) {
				throw new IndexException("Fehler beim Schreiben der Indexdatei '" + indexFile.getAbsolutePath() + "': " + e.getMessage(), e);
			}
		}
	}

		/**
			 * Schliet den Index und speichert alle genderten Daten auf Platte. Nach dem Aufruf von darf die Index-Instanz nicht mehr verwendet werden.
			 *
			 * @param basePath Verzeichnis, in dem die Index-Datei liegt
			 *
			 * @throws IndexException
			 */
	public void closeIndex(String basePath) throws IndexException {
		flush(basePath);
		byteArray = null;
	}

	/**
	 * Schliet die Indexdatei.
	 *
	 * @param indexFile Indexdatei
	 *
	 * @throws IndexException
	 */
	private void closeRandomAccessFile(RandomAccessFile indexFile) throws IndexException {
		try {
			if(indexFile != null) indexFile.close();
		}
		catch(IOException e) {
			throw new IndexException("Konnte die IndexDatei nicht schlieen. " + e.getMessage(), e);
		}
	}

	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("capacity=" + capacity() + ", entries=" + entries() + "\noldest= ");
		for(int i = 0; i < capacity(); i++) {
			if(rb2pos(rbFstUnsaved) == i) sb.append(" *unsaved* ==> ");
			sb.append(
					getEntry(DI, MIN, i) + "-" + getEntry(DI, MAX, i) + "/" + getEntry(AT, MIN, i) + "-" + getEntry(AT, MAX, i) + "/" + getEntry(DT, MIN, i)
					+ "-" + getEntry(DT, MAX, i) + "=" + getCID(i) + " "
			);
		}
		return sb.append("=newest").toString();
	}
}
